import tensorflow as tf

from DeepSaki.types.layers_enums import PaddingType,DownSampleType,UpSampleType,LinearLayerType
from DeepSaki.layers.sub_model_composites import Encoder, Decoder,Bottleneck
from DeepSaki.layers.layer_composites import Conv2DBlock, DenseBlock, ScalarGatedSelfAttention
from DeepSaki.layers.pooling import GlobalSumPooling2D

from typing import Tuple, Optional

class LayoutContentDiscriminator(tf.keras.Model):
    """Discriminator/critic model with two outputs to enforce disentanglement of layout and content discrimination.

    Usually used in a GAN framework.

    Info:
        Implementation as used in the [VoloGAN paper](https://arxiv.org/abs/2207.09204).

    Limitations:
        For now the input height and width of a tensor of shape (batch, height, width, channels) must be at least
        256x256

    Architecture:
        ```mermaid
        flowchart LR
        i([Input])-->e1
        subgraph Encoder
        e1-->e2-->e3-->sa[Self-Attention]
        end
        subgraph "Content Branch"
        sa-->c1-->c2-->c3-->c4
        end
        subgraph "Layout Branch"
        sa-->l1-->l2-->l3-->l4
        end
        l4-->lo([layout])
        c4-->co([content])
        ```

    **Example:**
    ```python
    import DeepSaki as ds
    import tensorflow as tf
    inputs = tf.keras.layers.Input(shape = (256,256,4))
    model = tf.keras.Model(inputs=inputs, outputs=ds.models.LayoutContentDiscriminator().call(inputs))
    model.summary()
    tf.keras.utils.plot_model(model, show_shapes=True, expand_nested=True, show_dtype=True, to_file='Unet_discriminator_model.png')
    ```
    """
    def __init__(self,
                filters:int = 64,
                downsampling:DownSampleType = DownSampleType.CONV_STRIDE_2,
                kernels:int = 3,
                first_kernel:int = 5,
                split_kernels:bool = False,
                number_of_blocks:int = 2,
                activation:str = "leaky_relu",
                dropout_rate:float = 0.2,
                use_spec_norm:bool=False,
                use_bias:bool = True,
                padding:PaddingType=PaddingType.ZERO,
                linear_layer_type:LinearLayerType = LinearLayerType.MLP,
                use_self_attention:bool = False,
                kernel_initializer:Optional[tf.keras.initializers.Initializer] = None,
                gamma_initializer:Optional[tf.keras.initializers.Initializer] = None
                ):
        """Initialize the `LayoutContentDiscriminator` object.

        Args:
            filters (int, optional): Number of filters for the initial encoder block. Defaults to 64.
            downsampling (DownSampleType, optional): Describes the downsampling method. Defaults to DownSampleType.CONV_STRIDE_2.
            kernels (int, optional): Size of the convolutions kernels. Defaults to 3.
            first_kernel (int, optional): The first convolution can have a different kernel size, to e.g. increase the
                perceptive field, while the channel depth is still low. Defaults to 5.
            split_kernels (bool, optional): To decrease the number of parameters, a convolution with the kernel_size
                `(kernel,kernel)` can be splitted into two consecutive convolutions with the kernel_size `(kernel,1)` and
                `(1,kernel)` respectivly. Defaults to False.
            number_of_blocks (int, optional): Number of consecutive convolutional building blocks, i.e. `Conv2DBlock`.
                Defaults to 2.
            activation (str, optional): String literal or tensorflow activation function object to obtain activation
                function. Defaults to "leaky_relu".
            dropout_rate (float, optional): Probability of the dropout layer. If the preceeding layer has more than one
                channel, spatial dropout is applied, otherwise standard dropout. Defaults to 0.2.
            use_spec_norm (bool, optional): Applies spectral normalization to convolutional and dense layers. Defaults to
                False.
            use_bias (bool, optional): Whether convolutions and dense layers include a bias or not. Defaults to True.
            padding (PaddingType, optional): Padding type. Defaults to PaddingType.NONE.
            linear_layer_type (LinearLayerType, optional): Determines whether 1x1 convolutions are replaced by linear
                layers, which gives the same result, but linear layers are faster. Option: LinearLayerType.MLP or
                LinearLayerType.CONV_1x1. Defaults to LinearLayerType.MLP.
            use_self_attention (bool, optional): Determines whether to apply self-attention in the decoder. Defaults to
                False.
            kernel_initializer (tf.keras.initializers.Initializer, optional): Initialization of the convolutions kernels.
                Defaults to None.
            gamma_initializer (tf.keras.initializers.Initializer, optional): Initialization of the normalization layers.
                Defaults to None.

        Raises:
            ValueError: If provided parameter for `linear_layer_type` is not supported.
        """
        super(LayoutContentDiscriminator, self).__init__()

        self.input_spec = tf.keras.layers.InputSpec(ndim=4)

        self.encoder = Encoder(number_of_levels=3, filters=filters, limit_filters=1024, downsampling=downsampling, kernels=kernels, split_kernels=split_kernels, number_of_blocks=number_of_blocks,activation=activation, first_kernel=first_kernel,use_ResidualBlock=False,channel_list=[4*filters,4*filters,8*filters],use_spec_norm=use_spec_norm, padding = padding, use_bias = use_bias, kernel_initializer = kernel_initializer, gamma_initializer = gamma_initializer)
        if use_self_attention:
            self.SA = ScalarGatedSelfAttention(use_spec_norm=use_spec_norm, intermediate_channel=None, kernel_initializer = kernel_initializer, gamma_initializer = gamma_initializer)
        else:
            self.SA = None

        if linear_layer_type == LinearLayerType.MLP:
            self.cont1 = DenseBlock(units = filters * 8, use_spec_norm = use_spec_norm, number_of_blocks = number_of_blocks, activation = activation, dropout_rate =dropout_rate, use_bias = use_bias, kernel_initializer = kernel_initializer, gamma_initializer = gamma_initializer)
            self.cont2 = DenseBlock(units = filters * 8, use_spec_norm = use_spec_norm, number_of_blocks = number_of_blocks, activation = activation, dropout_rate =dropout_rate, use_bias = use_bias, kernel_initializer = kernel_initializer, gamma_initializer = gamma_initializer)
            self.cont3 = DenseBlock(units = filters * 8, use_spec_norm = use_spec_norm, number_of_blocks = number_of_blocks, activation = activation, dropout_rate =0, final_activation=False, apply_final_normalization = False, use_bias = use_bias, kernel_initializer = kernel_initializer, gamma_initializer = gamma_initializer)
        elif linear_layer_type == LinearLayerType.CONV_1x1:
            self.cont1 = Conv2DBlock(filters=filters * 8, kernels = 1, activation = activation, split_kernels = split_kernels,number_of_blocks=number_of_blocks, dropout_rate=dropout_rate,use_spec_norm=use_spec_norm, padding=padding,use_bias = use_bias, kernel_initializer = kernel_initializer, gamma_initializer = gamma_initializer)
            self.cont2 = Conv2DBlock(filters=filters * 8, kernels = 1, activation = activation, split_kernels = split_kernels,number_of_blocks=number_of_blocks, dropout_rate=dropout_rate,use_spec_norm=use_spec_norm, padding=padding,use_bias = use_bias, kernel_initializer = kernel_initializer, gamma_initializer = gamma_initializer)
            self.cont3 = Conv2DBlock(filters=filters * 8, kernels = 1, activation = activation, split_kernels = split_kernels,number_of_blocks=number_of_blocks, dropout_rate=0,use_spec_norm=use_spec_norm, final_activation=False, apply_final_normalization = False, padding=padding,use_bias = use_bias, kernel_initializer = kernel_initializer, gamma_initializer = gamma_initializer)
        else:
            raise ValueError(f"{linear_layer_type=} is not defined.")

        self.lay1 = Conv2DBlock(filters=1, kernels = kernels, activation = activation, split_kernels = split_kernels,number_of_blocks=number_of_blocks, dropout_rate=0,use_spec_norm=use_spec_norm, padding=padding, use_bias = use_bias, kernel_initializer = kernel_initializer, gamma_initializer = gamma_initializer)
        self.lay2 = Conv2DBlock(filters=1, kernels = kernels, activation = activation, split_kernels = split_kernels,number_of_blocks=number_of_blocks, dropout_rate=dropout_rate,use_spec_norm=use_spec_norm, padding=padding,use_bias = use_bias, kernel_initializer = kernel_initializer, gamma_initializer = gamma_initializer)
        self.lay3 = Conv2DBlock(filters=1, kernels = kernels, activation = activation, split_kernels = split_kernels,number_of_blocks=number_of_blocks, dropout_rate=dropout_rate,use_spec_norm=use_spec_norm, padding=padding,use_bias = use_bias, kernel_initializer = kernel_initializer, gamma_initializer = gamma_initializer)
        self.lay4 = Conv2DBlock(filters=1, kernels = kernels, activation = activation, split_kernels = split_kernels,number_of_blocks=number_of_blocks, dropout_rate=0,use_spec_norm=use_spec_norm,final_activation=False, apply_final_normalization = False, padding=padding,use_bias = use_bias, kernel_initializer = kernel_initializer, gamma_initializer = gamma_initializer)
        #To enable mixed precission support for matplotlib and distributed training and to increase training stability
        self.linear_dtype = tf.keras.layers.Activation("linear", dtype = tf.float32)

    def build(self, input_shape:tf.TensorShape)->None:
        _, height, width,_ = input_shape
        if height < 256 or width < 256:
            raise ValueError(f"Input requires a height and width of minimum 256, but got {height=} and {width=}")

    def call(self, inputs:tf.Tensor)->Tuple[tf.Tensor,tf.Tensor]:
        """Calls the `LayoutContentDiscriminator` model.

        Args:
            inputs (tf.Tensor): Tensor of shape `(batch, height, width, channel)`.

        Returns:
            content_output: Tensor of shape `(batch, 1, 1, C)` focusing on the content of an image rather than the position.
                `C` is `filters`*2**3, where 3 is the number of downsampling blocks in the Encoder().
            layout_output: Tensor of shape `(batch, H, W, 1)` focusing on the input's layout by reducing its number of
                channels to 1. `H`=`height`/(2^3) and `W`=`width`/(2^3)
        """
        x = inputs
        encoder_out = self.encoder(x)
        if self.SA is not None:
            encoder_out = self.SA(encoder_out)

        _, pool_height, pool_width, _ = encoder_out.shape
        out1 = tf.keras.layers.AveragePooling2D(pool_size=(pool_height, pool_width))(encoder_out)
        out1 = self.cont1(out1)
        out1 = self.cont2(out1)
        out1 = self.cont3(out1)
        out1 = self.linear_dtype(out1)

        out2 = self.lay1(encoder_out)
        out2 = self.lay2(out2)
        out2 = self.lay3(out2)
        out2 = self.lay4(out2)
        out2 = self.linear_dtype(out2)

        return out1, out2

class PatchDiscriminator(tf.keras.Model):
    """Discriminator/critic model with patched output rather than single value.

    Usually used in a GAN framework.

    Architecture:
        ```mermaid
        flowchart LR
        i([Input])-->e1
        subgraph Encoder
        e1-->e2-->ex
        end
        ex-->conv1-->conv2-->o([output])

        ```

    **Example:**
    ```python
    import DeepSaki as ds
    import tensorflow as tf
    inputs = tf.keras.layers.Input(shape = (256,256,4))
    model = tf.keras.Model(inputs=inputs, outputs=ds.models.PatchDiscriminator().call(inputs))
    model.summary()
    tf.keras.utils.plot_model(model, show_shapes=True, expand_nested=True, show_dtype=True, to_file='PatchDiscriminator_model.png')
    ```

    """
    def __init__(self,
                filters:int = 64,
                downsampling:DownSampleType = DownSampleType.AVG_POOLING,
                kernels:int = 3,
                first_kernel:int = 5,
                split_kernels:bool = False,
                number_of_blocks:int = 2,
                activation:str = "leaky_relu",
                num_down_blocks:int = 3,
                dropout_rate:float = 0.2,
                use_spec_norm:bool= False,
                use_bias:bool = True,
                use_self_attention:bool=False,
                padding:PaddingType=PaddingType.ZERO,
                kernel_initializer:Optional[tf.keras.initializers.Initializer] = None,
                gamma_initializer:Optional[tf.keras.initializers.Initializer] = None
                ):
        """Initialize the `PatchDiscriminator` object.

        Args:
            filters (int, optional): Number of filters for the initial encoder block. Defaults to 64.
            downsampling (DownSampleType, optional): Describes the downsampling method. Defaults to DownSampleType.AVG_POOLING.
            kernels (int, optional): Size of the convolutions kernels. Defaults to 3.
            first_kernel (int, optional): The first convolution can have a different kernel size, to e.g. increase the
                perceptive field, while the channel depth is still low. Defaults to 5.
            split_kernels (bool, optional): To decrease the number of parameters, a convolution with the kernel_size
                `(kernel,kernel)` can be splitted into two consecutive convolutions with the kernel_size `(kernel,1)` and
                `(1,kernel)` respectivly. Defaults to False.
            number_of_blocks (int, optional): Number of consecutive convolutional building blocks, i.e. `Conv2DBlock`.
                Defaults to 2.
            activation (str, optional): String literal or tensorflow activation function object to obtain activation
                function. Defaults to "leaky_relu".
            num_down_blocks (int, optional): Number of levels in the encoder that performs a downsampling operation at its
                end. Defaults to 3.
            dropout_rate (float, optional): Probability of the dropout layer. If the preceeding layer has more than one
                channel, spatial dropout is applied, otherwise standard dropout. Defaults to 0.2.
            use_spec_norm (bool, optional): Applies spectral normalization to convolutional and dense layers. Defaults to
                False.
            use_bias (bool, optional): Whether convolutions and dense layers include a bias or not. Defaults to True.
            use_self_attention (bool, optional): Determines whether to apply self-attention in the decoder. Defaults to
                False.
            padding (PaddingType, optional): Padding type. Defaults to PaddingType.NONE.
            kernel_initializer (tf.keras.initializers.Initializer, optional): Initialization of the convolutions kernels.
                Defaults to None.
            gamma_initializer (tf.keras.initializers.Initializer, optional): Initialization of the normalization layers.
                Defaults to None.
        """
        super(PatchDiscriminator, self).__init__()

        self.input_spec = tf.keras.layers.InputSpec(ndim=4)

        self.encoder = Encoder(number_of_levels=num_down_blocks, filters=filters, limit_filters=512,  downsampling=downsampling, kernels=kernels, split_kernels=split_kernels,number_of_blocks=number_of_blocks,activation=activation,first_kernel=first_kernel,use_spec_norm=use_spec_norm,use_self_attention=use_self_attention, use_bias = use_bias,padding = padding, kernel_initializer = kernel_initializer, gamma_initializer = gamma_initializer)
        self.conv1 = Conv2DBlock(filters = filters * (2**(num_down_blocks)),  kernels = kernels, split_kernels = split_kernels, number_of_blocks = number_of_blocks, activation = activation,dropout_rate=dropout_rate,use_spec_norm=use_spec_norm, use_bias = use_bias,padding = padding, kernel_initializer = kernel_initializer, gamma_initializer = gamma_initializer)
        self.conv2 = Conv2DBlock(filters = 1, kernels = 5, split_kernels  = False, number_of_blocks = 1, activation = None,use_spec_norm=use_spec_norm, apply_final_normalization = False, use_bias = use_bias,padding = padding, kernel_initializer = kernel_initializer, gamma_initializer = gamma_initializer)
        #To enable mixed precission support for matplotlib and distributed training and to increase training stability
        self.linear_dtype = tf.keras.layers.Activation("linear", dtype = tf.float32)

    def call(self, inputs:tf.Tensor)->tf.Tensor:
        """Calls the `PatchDiscriminator` model.

        Args:
            inputs (tf.Tensor): Tensor of shape `(batch, height, width, channel)`.

        Returns:
            tf.Tensor: Tensor of shape `(batch, height//2**`number_of_blocks`, width//2**`number_of_blocks`, 1)`.
        """
        x = inputs
        x = self.encoder(x)
        x = self.conv1(x)
        x = self.conv2(x)
        return self.linear_dtype(x)

class UNetDiscriminator(tf.keras.Model):
    """U-Net based discriminator for pixel-wise real/fake prediction plus additional output for global prediction.

    Usually used in a GAN framework.

    Info:
        Implementation as used in the [VoloGAN paper](https://arxiv.org/abs/2207.09204).

    Architecture:
        ```mermaid
        flowchart LR
        i([Input])-->e1
        subgraph Encoder
        e1-->e2-->e3-->ex
        end
        subgraph Bottleneck
        ex-->b1-->bx
        end
        bx-->gp[GlobalSumPooling]-->sn[SPecNorm]-->eo([Global Prediction])
        subgraph Decoder
        bx-->dx-->d3-->d2-->d1
        end
        d1-->o([Pixel-Wise Prediction])
        e1-->d1
        e2-->d2
        e3-->d3
        ex-->dx
        ```

    **Example:**
    ```python
    import DeepSaki as ds
    import tensorflow as tf
    inputs = tf.keras.layers.Input(shape = (256,256,4))
    model = tf.keras.Model(inputs=inputs, outputs=ds.models.UNetDiscriminator(5).call(inputs))
    model.summary()
    tf.keras.utils.plot_model(model, show_shapes=True, expand_nested=True, show_dtype=True, to_file='model.png')
    ```
    """
    def __init__(self,
                number_of_levels:int= 4,
                upsampling:UpSampleType = UpSampleType.TRANSPOSE_CONV,
                downsampling:DownSampleType = DownSampleType.AVG_POOLING,
                kernels:int = 3,
                first_kernel:int = 5,
                split_kernels:bool = False,
                number_of_blocks:int = 2,
                activation:str = "leaky_relu",
                use_ResidualBlock:bool = False,
                residual_cardinality:int = 1,
                limit_filters:int = 512,
                n_bottleneck_blocks:int = 1,
                filters:int = 64,
                dropout_rate:float = 0.2,
                use_self_attention:bool=False,
                use_spec_norm:bool=False,
                use_bias:bool= True,
                linear_layer_type:LinearLayerType = LinearLayerType.MLP,
                padding:PaddingType=PaddingType.ZERO,
                kernel_initializer:Optional[tf.keras.initializers.Initializer] = None,
                gamma_initializer:Optional[tf.keras.initializers.Initializer] =  None
                ):
        """Initialize the `UNetDiscriminator` object.

        Args:
            number_of_levels (int, optional): Number of down and upsampling levels of the model. Defaults to 4.
            upsampling (UpSampleType, optional): Describes the upsampling method used. Defaults to UpSampleType.TRANSPOSE_CONV.
            downsampling (DownSampleType, optional): Describes the downsampling method. Defaults to DownSampleType.AVG_POOLING.
            kernels (int, optional): Size of the convolutions kernels. Defaults to 3.
            first_kernel (int, optional): The first convolution can have a different kernel size, to e.g. increase the
                perceptive field, while the channel depth is still low. Defaults to 5.
            split_kernels (bool, optional): To decrease the number of parameters, a convolution with the kernel_size
                `(kernel,kernel)` can be splitted into two consecutive convolutions with the kernel_size `(kernel,1)` and
                `(1,kernel)` respectivly. Defaults to False.
            number_of_blocks (int, optional): Number of consecutive convolutional building blocks, i.e. `Conv2DBlock`.
                Defaults to 2.
            activation (str, optional): String literal or tensorflow activation function object to obtain activation
                function. Defaults to "leaky_relu".
            use_ResidualBlock (bool, optional): Whether or not to use the ResidualBlock instead of the
                `Conv2DBlock`. Defaults to False.
            residual_cardinality (int, optional): Cardinality for the `ResidualBlock`. Defaults to 1.
            limit_filters (int, optional): Limits the number of filters, which is doubled with every downsampling block.
                Defaults to 512.
            n_bottleneck_blocks (int, optional): Number of consecutive convolution blocks in the bottleneck. Defaults to 1.
            filters (int, optional): Number of filters for the initial encoder block. Defaults to 64.
            dropout_rate (float, optional): Probability of the dropout layer. If the preceeding layer has more than one
                channel, spatial dropout is applied, otherwise standard dropout. Defaults to 0.2.
            use_self_attention (bool, optional): Determines whether to apply self-attention in the decoder. Defaults to
                False.
            use_spec_norm (bool, optional): Applies spectral normalization to convolutional and dense layers. Defaults to
                False.
            use_bias (bool, optional): Whether convolutions and dense layers include a bias or not. Defaults to True.
            linear_layer_type (LinearLayerType, optional): Determines whether 1x1 convolutions are replaced by linear
                layers, which gives the same result, but linear layers are faster. Option: LinearLayerType.MLP or
                LinearLayerType.CONV_1x1. Defaults to LinearLayerType.MLP.
            padding (PaddingType, optional): Padding type. Defaults to PaddingType.ZERO.
            kernel_initializer (tf.keras.initializers.Initializer, optional): Initialization of the convolutions kernels.
                Defaults to None.
            gamma_initializer (tf.keras.initializers.Initializer, optional): Initialization of the normalization layers.
                Defaults to None.
        """
        super(UNetDiscriminator, self).__init__()
        ch = filters

        self.input_spec = tf.keras.layers.InputSpec(ndim=4)

        self.encoder = Encoder(number_of_levels=number_of_levels, filters=filters, limit_filters=limit_filters,  downsampling=downsampling, kernels=kernels, split_kernels=split_kernels, number_of_blocks=number_of_blocks,activation=activation, first_kernel=first_kernel, use_ResidualBlock=use_ResidualBlock, channel_list=[ch,2*ch,4*ch,8*ch,8*ch], use_spec_norm=use_spec_norm,use_self_attention=use_self_attention,output_skips=True, use_bias = use_bias,residual_cardinality=residual_cardinality,padding = padding, kernel_initializer = kernel_initializer, gamma_initializer = gamma_initializer)
        self.bottleNeck = Bottleneck(use_ResidualBlock=use_ResidualBlock, n_bottleneck_blocks=n_bottleneck_blocks, kernels=kernels, split_kernels=split_kernels,number_of_blocks=number_of_blocks, activation = activation,dropout_rate=dropout_rate, channel_list=[16*ch], use_spec_norm=use_spec_norm, use_bias = use_bias,residual_cardinality=residual_cardinality,padding = padding, kernel_initializer = kernel_initializer, gamma_initializer = gamma_initializer)
        self.decoder = Decoder(number_of_levels=number_of_levels, upsampling=upsampling, filters=filters, limit_filters=limit_filters, kernels=kernels, split_kernels=split_kernels,number_of_blocks=number_of_blocks,activation=activation,dropout_rate=dropout_rate, use_ResidualBlock=use_ResidualBlock, channel_list=[8*ch,8*ch,4*ch,2*ch,ch], use_spec_norm=use_spec_norm, use_bias = use_bias,residual_cardinality=residual_cardinality,padding = padding, kernel_initializer = kernel_initializer, gamma_initializer = gamma_initializer,enable_skip_connections_input=True)
        if linear_layer_type == LinearLayerType.MLP:
            self.img_reconstruction = DenseBlock(units = 1, use_spec_norm = use_spec_norm, number_of_blocks = 1, activation = None, apply_final_normalization = False, use_bias = use_bias, kernel_initializer = kernel_initializer, gamma_initializer = gamma_initializer)
        elif linear_layer_type == LinearLayerType.CONV_1x1:
            self.img_reconstruction = Conv2DBlock(filters = 1,  kernels = 1, split_kernels  = False, number_of_blocks = 1, activation = None,use_spec_norm=use_spec_norm, apply_final_normalization = False, use_bias = use_bias,padding = padding, kernel_initializer = kernel_initializer, gamma_initializer = gamma_initializer)
        self.linear = DenseBlock(units = 1, use_spec_norm = use_spec_norm, number_of_blocks = 1, activation = None, apply_final_normalization = False, use_bias = use_bias, kernel_initializer = kernel_initializer, gamma_initializer = gamma_initializer)
        #To enable mixed precission support for matplotlib and distributed training and to increase training stability
        self.linear_dtype = tf.keras.layers.Activation("linear", dtype = tf.float32)


    def call(self, inputs:tf.Tensor)->Tuple[tf.Tensor,tf.Tensor]:
        """Calls the `UNetDiscriminator` model.

        Args:
            inputs (tf.Tensor): Tensor of shape (`batch`,`height`,`width`,`channel`).

        Returns:
            decoder_output: Real/fake prediction for each pixel of shape (batch, height, width, 1)
            encoder_output: Global real/fake prediction of shape (batch, 1).
        """
        x = inputs
        encoder_output, skip_connections = self.encoder(x)
        bottleneck_output = self.bottleNeck(encoder_output)
        decoder_output = self.decoder([bottleneck_output,skip_connections])

        out1 = GlobalSumPooling2D()(bottleneck_output)
        out1 = self.linear(out1)
        out1 = self.linear_dtype(out1)

        out2 = self.img_reconstruction(decoder_output)
        out2 = self.linear_dtype(out2)
        return out1, out2



import inspect
import os

import pytest

os.environ["TF_CPP_MIN_LOG_LEVEL"] = "2"  # deactivate tensorflow warnings and infos. Keep Errors

from tensorflow.python.keras.optimizer_v2 import optimizer_v2

from DeepSaki.optimizers.swats import SwatsAdam
from DeepSaki.optimizers.swats import SwatsNadam


@pytest.mark.parametrize(
    "optimizer_object",
    [
        SwatsAdam(),
        SwatsNadam(),
    ],
)
class TestGenericOptimizer:
    def test_get_dict(self, optimizer_object):
        config = optimizer_object.get_config()
        # func bellow gets all variable names of the __init__ param list. [1::] removes "self" from that list.
        expected_keys = inspect.getfullargspec(optimizer_object.__init__)[0][1::]
        key_in_config = [key in config for key in expected_keys]
        assert all(key_in_config), f"not all expected keys found in config: {key_in_config}"

    def test_layer_is_subclass_of_tensorflow_layer(self, optimizer_object):
        assert isinstance(optimizer_object, optimizer_v2.OptimizerV2)



import logging
from typing import List
from typing import Optional
from typing import Tuple
from typing import Union

import tensorflow as tf
from tensorflow.python.client import device_lib

def detect_accelerator(
    gpu_memory_groth: bool = False,
) -> Tuple[
    tf.distribute.Strategy,
    str,
    Optional[Union[tf.distribute.cluster_resolver.TPUClusterResolver, List[tf.config.LogicalDevice]]],
]:
    """Detects the availability of TPUs and GPUs and connects to all available devices.

    Args:
        gpu_memory_groth (bool, optional): If true, memory is allocated on demand. Defaults to False.

    Returns:
        `strategy`: pointer to the distribution strategy configuration object
        `runtime_environment`: Info string describing the HW that might be used for conditions, i.e. "TPU, "GPU", "CPU"
        `hw_accelerator_handle`: pointer to the HW accelerator
    """
    hw_accelerator_handle = None
    tpu = None
    gpus = []
    try:
        tpu = tf.distribute.cluster_resolver.TPUClusterResolver()  # TPU detection
        hw_accelerator_handle = tpu
    except ValueError:
        if gpu_memory_groth:
            for gpu in tf.config.experimental.list_physical_devices("GPU"):
                tf.config.experimental.set_memory_growth(gpu, enable=True)
        gpus = tf.config.experimental.list_logical_devices("GPU")
        hw_accelerator_handle = gpus

    # Select appropriate distribution strategy
    if tpu is not None:
        runtime_environment = "TPU"
        tf.config.experimental_connect_to_cluster(tpu)
        tf.tpu.experimental.initialize_tpu_system(tpu)
        strategy = tf.distribute.TPUStrategy(
            tpu
        )  # Going back and forth between TPU and host is expensive. Better to run 128 batches on the TPU before reporting back.
        logging.info(f"Running on TPU {tpu.cluster_spec().as_dict()['worker']}")
    elif len(gpus) > 1:
        runtime_environment = "GPU"
        strategy = tf.distribute.MirroredStrategy([gpu.name for gpu in gpus])
        logging.info(f"Running on multiple GPUs {[gpu.name for gpu in gpus]}")
    elif len(gpus) == 1:
        runtime_environment = "GPU"
        strategy = tf.distribute.get_strategy()  # default strategy that works on CPU and single GPU
        logging.info(f"Running on single GPU {gpus[0].name}")
    else:
        runtime_environment = "CPU"
        strategy = tf.distribute.get_strategy()  # default strategy that works on CPU and single GPU
        logging.info("Running on CPU")

    logging.info(f"Number of accelerators: {strategy.num_replicas_in_sync}")
    logging.info("____________________________________________________________________________________")
    logging.info("Device List: ")
    logging.info(device_lib.list_local_devices())

    return strategy, runtime_environment, hw_accelerator_handle


def enable_xla_acceleration() -> None:
    """Enable compiler acceleration for linear algebra operations."""
    tf.config.optimizer.set_jit(enabled="autoclustering")
    logging.info("Linear algebra acceleration enabled")


def enable_mixed_precision() -> None:
    """Set mixed precission policy depending on the available HW accelerator. TPU:`mixed_bfloat16`. GPU:`mixed_float16`."""
    try:
        tpu = tf.distribute.cluster_resolver.TPUClusterResolver()  # TPU detection
    except ValueError:
        tpu = None

    policy_config = "mixed_bfloat16" if tpu else "mixed_float16"
    policy = tf.keras.mixed_precision.Policy(policy_config)
    tf.keras.mixed_precision.set_global_policy(policy)
    logging.info(f"Mixed precision enabled to {policy_config}")



import os

import pytest

os.environ["TF_CPP_MIN_LOG_LEVEL"] = "2"  # deactivate tensorflow warnings and infos. Keep Errors
import tensorflow as tf

from DeepSaki.initializers.he_alpha import HeAlphaNormal
from DeepSaki.initializers.he_alpha import HeAlphaUniform
from DeepSaki.layers.layer_helper import InitializerFunc
from DeepSaki.types.layers_enums import PaddingType
from DeepSaki.layers.layer_helper import dropout_func
from DeepSaki.layers.layer_helper import get_initializer
from DeepSaki.layers.layer_helper import pad_func
from DeepSaki.layers.padding import ReflectionPadding2D


class TestGetInitializerFunc:
    @pytest.mark.parametrize(
        ("initializer_func", "expected"),
        [
            (InitializerFunc.RANDOM_NORMAL, tf.keras.initializers.RandomNormal),
            (InitializerFunc.RANDOM_UNIFORM, tf.keras.initializers.RandomUniform),
            (InitializerFunc.GLOROT_NORMAL, tf.keras.initializers.GlorotNormal),
            (InitializerFunc.GLOROT_UNIFORM, tf.keras.initializers.GlorotUniform),
            (InitializerFunc.HE_NORMAL, tf.keras.initializers.HeNormal),
            (InitializerFunc.HE_UNIFORM, tf.keras.initializers.HeUniform),
            (InitializerFunc.HE_ALPHA_NORMAL, HeAlphaNormal),
            (InitializerFunc.HE_ALPHA_UNIFORM, HeAlphaUniform),
        ],
    )
    def test_get_initializer_func_returns_correct_initializer(self, initializer_func, expected):
        returned_initializer = get_initializer(initializer_func)
        assert isinstance(returned_initializer, expected)

    @pytest.mark.parametrize(
        "initializer_func",
        [InitializerFunc.NONE, None, "Any_string"],
    )
    def test_get_initializer_func_raises_value_error(self, initializer_func):
        with pytest.raises(ValueError):
            _ = get_initializer(initializer_func)


class TestPadFunc:
    @pytest.mark.parametrize(
        ("padding_type", "expected"),
        [
            (PaddingType.ZERO, tf.keras.layers.ZeroPadding2D),
            (PaddingType.REFLECTION, ReflectionPadding2D),
        ],
    )
    def test_pad_func_returns_correct_padding(self, padding_type, expected):
        returned_padding = pad_func(padding_type=padding_type)
        assert isinstance(returned_padding, expected)

    @pytest.mark.parametrize(
        "padding_type",
        [PaddingType.NONE, None, "Any_string"],
    )
    def test_pad_func_raises_value_error(self, padding_type):
        with pytest.raises(ValueError):
            _ = pad_func(padding_type=padding_type)


@pytest.mark.parametrize(
    ("input_filters", "expected"),
    [
        (1, tf.keras.layers.Dropout),
        (2, tf.keras.layers.SpatialDropout2D),
        (64, tf.keras.layers.SpatialDropout2D),
        (99, tf.keras.layers.SpatialDropout2D),
    ],
)
def test_dropout_func_returns_correct_dropout(input_filters, expected):
    output = dropout_func(filters=input_filters, dropout_rate=0.5)
    assert isinstance(output, expected)


@pytest.mark.parametrize(
    ("input_filters", "error_type"),
    [
        (0, ValueError),
        (-10, ValueError),
        (-99, ValueError),
        ("Any String", TypeError),
        (None, TypeError),
        (3.0, TypeError),
    ],
)
def test_dropout_func_raises_error(input_filters, error_type):
    with pytest.raises(error_type):
        _ = dropout_func(filters=input_filters, dropout_rate=0.5)


@pytest.mark.skip(reason="Not implemented yet.")
def test_plot_layer():
    ...



import os

import pytest

os.environ["TF_CPP_MIN_LOG_LEVEL"] = "2"  # deactivate tensorflow warnings and infos. Keep Errors
import tensorflow as tf

from DeepSaki.initializers.complex_initializer import ComplexInitializer


@pytest.mark.parametrize(
    ("initializer_imag"),
    [
        None,
        tf.initializers.HeNormal(seed=123),
        tf.keras.initializers.RandomNormal(seed=123),
        tf.initializers.GlorotUniform(seed=123),
    ],
)
@pytest.mark.parametrize(
    ("initializer_real"),
    [
        (tf.initializers.HeNormal(seed=456)),
        (tf.keras.initializers.RandomNormal(seed=456)),
        (tf.initializers.GlorotUniform(seed=456)),
    ],
)
@pytest.mark.parametrize("dtype", [tf.complex64, tf.complex128])
@pytest.mark.parametrize("shape", [(2, 2), (3, 2, 1), (1, 2, 3, 4)])
class TestComplexInitializer:
    def _get_initializer_values(self, initializer_real, initializer_imag, dtype, shape):
        complex_initializer = ComplexInitializer(initializer_real, initializer_imag)
        return complex_initializer(shape, dtype)

    def test_initializer_values_complex(self, initializer_real, initializer_imag, dtype, shape):
        values = self._get_initializer_values(initializer_real, initializer_imag, dtype, shape)
        assert values.dtype == dtype

    def test_initializer_values_correct_shape(self, initializer_real, initializer_imag, dtype, shape):
        values = self._get_initializer_values(initializer_real, initializer_imag, dtype, shape)
        assert values.shape == shape

    def test_real_imag_same_if_single_initializer(self, initializer_real, initializer_imag, dtype, shape):
        values = self._get_initializer_values(initializer_real, initializer_imag, dtype, shape)

        assert tf.math.reduce_all(
            (tf.math.real(values).numpy() == pytest.approx(tf.math.imag(values).numpy(), 0.01))
            == (initializer_imag is None)
        )



# he_alpha.py
from DeepSaki.initializers.he_alpha import HeAlpha
from DeepSaki.initializers.he_alpha import HeAlphaUniform
from DeepSaki.initializers.he_alpha import HeAlphaNormal

# initializer_helper.py
from DeepSaki.initializers.complex_initializer import ComplexInitializer

__all__ = [
    "HeAlpha",
    "HeAlphaUniform",
    "HeAlphaNormal",
    "ComplexInitializer",
]



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from typing import Any
from typing import Dict
from typing import List

import tensorflow as tf
from tensorflow.python.framework import ops
from tensorflow.python.keras import backend_config
from tensorflow.python.keras.optimizer_v2 import optimizer_v2
from tensorflow.python.ops import array_ops
from tensorflow.python.ops import control_flow_ops
from tensorflow.python.ops import math_ops
from tensorflow.python.ops import state_ops
from tensorflow.python.ops import variables as tf_variables
from tensorflow.python.training import gen_training_ops

from DeepSaki.types.optimizers_enums import CurrentOptimizer

class SwatsAdam(optimizer_v2.OptimizerV2):
    """Initializer that can switch from ADAM to SGD and vice versa.

    SWATS_ADAM is inspired by the [SWATS](http://arxiv.org/abs/1712.07628) (switching from adam to sgd) initializer.
    During training, the optimizer can be changed by calling `switch_optimizer()` uppon the optimizer instance.

    This optimizer combines tensorflow's [ADAM](https://github.com/tensorflow/tensorflow/blob/85c8b2a817f95a3e979ecd1ed95bff1dc1335cff/tensorflow/python/keras/optimizer_v2/adam.py)
    and [SGD](https://github.com/tensorflow/tensorflow/blob/85c8b2a817f95a3e979ecd1ed95bff1dc1335cff/tensorflow/python/keras/optimizer_v2/gradient_descent.py) implementations.

    **Example:**
    ```python
    import DeepSaki as ds
    optimizer = ds.optimizer.SWATS_ADAM(learning_rate=disc_learning_rate, beta_1=0.5, amsgrad=False, momentum=0.9, nesterov=False)
    optimizer.switch_optimizer()
    ```
    """

    _HAS_AGGREGATE_GRAD = True

    def __init__(
        self,
        learning_rate: float = 0.001,
        beta_1: float = 0.9,
        beta_2: float = 0.999,
        epsilon: float = 1e-7,
        amsgrad: bool = False,
        momentum: float = 0.0,
        nesterov: bool = False,
        **kwargs: Any,
    ) -> None:
        """Initializes the `SwatsAdam` optimizer.

        Args:
            learning_rate (float, optional): Learning rate of the optimization step. Defaults to 0.001.
            beta_1 (float, optional): The exponential decay rate for the 1st moment estimates. Defaults to 0.9.
            beta_2 (float, optional): The exponential decay rate for the 2nd moment estimates. Defaults to 0.999.
            epsilon (float, optional): A small constant for numerical stability. Defaults to 1e-7.
            amsgrad (bool, optional): Whether to apply AMSGrad variant of this algorithm from the paper "On the
                Convergence of Adam and beyond". Defaults to False.
            momentum (float, optional):  Hyperparameter >= 0 that accelerates gradient descent in the relevant direction
                and dampens oscillations. Vanilla gradient descent means no momentum. Defaults to 0.0.
            nesterov (bool, optional): Whether to apply Nesterov momentum. Defaults to False.
            kwargs: keyword arguments passed to the parent class.

        Raises:
            ValueError: if `momentum` is not provided in an expected form.
        """
        super(SwatsAdam, self).__init__(name="SwatsAdam", **kwargs)
        self._set_hyper("learning_rate", kwargs.get("lr", learning_rate))
        self._set_hyper("decay", self._initial_decay)
        self._set_hyper("beta_1", beta_1)
        self._set_hyper("beta_2", beta_2)
        self.epsilon = epsilon or backend_config.epsilon()
        self.amsgrad = amsgrad
        self.currentOptimizer = CurrentOptimizer.ADAM

        self._momentum = False
        if isinstance(momentum, ops.Tensor) or callable(momentum) or momentum > 0:
            self._momentum = True
        if isinstance(momentum, (int, float)) and (momentum < 0 or momentum > 1):
            raise ValueError("`momentum` must be between [0, 1].")
        self._set_hyper("momentum", momentum)

        self.nesterov = nesterov

    def switch_optimizer(self) -> None:
        """Switches the current optimizer to either ADAM or SGD, depending on what was previously set."""
        if self.currentOptimizer == CurrentOptimizer.ADAM:
            self.currentOptimizer = CurrentOptimizer.SGD
        else:
            self.currentOptimizer = CurrentOptimizer.ADAM

    def _create_slots(self, var_list: List[tf.Variable]) -> None:
        # Create slots for the first and second moments.
        # Separate for-loops to respect the ordering of slot variables from v1.
        for var in var_list:
            self.add_slot(var, "m")
        for var in var_list:
            self.add_slot(var, "v")
        if self.amsgrad:
            for var in var_list:
                self.add_slot(var, "vhat")
        if self._momentum:
            for var in var_list:
                self.add_slot(var, "momentum")

    def _prepare_local(self, var_device, var_dtype: tf.DType, apply_state) -> None:
        super(SwatsAdam, self)._prepare_local(var_device, var_dtype, apply_state)

        local_step = math_ops.cast(self.iterations + 1, var_dtype)
        beta_1_t = array_ops.identity(self._get_hyper("beta_1", var_dtype))
        beta_2_t = array_ops.identity(self._get_hyper("beta_2", var_dtype))
        beta_1_power = math_ops.pow(beta_1_t, local_step)
        beta_2_power = math_ops.pow(beta_2_t, local_step)
        lr = apply_state[(var_device, var_dtype)]["lr_t"] * (math_ops.sqrt(1 - beta_2_power) / (1 - beta_1_power))
        apply_state[(var_device, var_dtype)].update(
            {
                "lr": lr,
                "epsilon": ops.convert_to_tensor_v2_with_dispatch(self.epsilon, var_dtype),
                "beta_1_t": beta_1_t,
                "beta_1_power": beta_1_power,
                "one_minus_beta_1_t": 1 - beta_1_t,
                "beta_2_t": beta_2_t,
                "beta_2_power": beta_2_power,
                "one_minus_beta_2_t": 1 - beta_2_t,
            }
        )

        apply_state[(var_device, var_dtype)]["momentum"] = array_ops.identity(self._get_hyper("momentum", var_dtype))

    def set_weights(self, weights) -> None:
        params = self.weights
        # If the weights are generated by Keras V1 optimizer, it includes vhats
        # even without amsgrad, i.e, V1 optimizer has 3x + 1 variables, while V2
        # optimizer has 2x + 1 variables. Filter vhats out for compatibility.
        num_vars = int((len(params) - 1) / 2)
        if len(weights) == 3 * num_vars + 1:
            weights = weights[: len(params)]
        super(SwatsAdam, self).set_weights(weights)

    def _resource_apply_dense(self, grad, var, apply_state=None):
        var_device, var_dtype = var.device, var.dtype.base_dtype
        coefficients = (apply_state or {}).get((var_device, var_dtype)) or self._fallback_apply_state(
            var_device, var_dtype
        )

        if self.currentOptimizer == CurrentOptimizer.SGD:
            if self._momentum:
                momentum_var = self.get_slot(var, "momentum")
                return gen_training_ops.ResourceApplyKerasMomentum(
                    var=var.handle,
                    accum=momentum_var.handle,
                    lr=coefficients["lr_t"],
                    grad=grad,
                    momentum=coefficients["momentum"],
                    use_locking=self._use_locking,
                    use_nesterov=self.nesterov,
                )

            return gen_training_ops.ResourceApplyGradientDescent(
                var=var.handle, alpha=coefficients["lr_t"], delta=grad, use_locking=self._use_locking
            )

        if self.currentOptimizer == CurrentOptimizer.ADAM:
            m = self.get_slot(var, "m")
            v = self.get_slot(var, "v")
            if not self.amsgrad:
                return gen_training_ops.ResourceApplyAdam(
                    var=var.handle,
                    m=m.handle,
                    v=v.handle,
                    beta1_power=coefficients["beta_1_power"],
                    beta2_power=coefficients["beta_2_power"],
                    lr=coefficients["lr_t"],
                    beta1=coefficients["beta_1_t"],
                    beta2=coefficients["beta_2_t"],
                    epsilon=coefficients["epsilon"],
                    grad=grad,
                    use_locking=self._use_locking,
                )

            vhat = self.get_slot(var, "vhat")
            return gen_training_ops.ResourceApplyAdamWithAmsgrad(
                var=var.handle,
                m=m.handle,
                v=v.handle,
                vhat=vhat.handle,
                beta1_power=coefficients["beta_1_power"],
                beta2_power=coefficients["beta_2_power"],
                lr=coefficients["lr_t"],
                beta1=coefficients["beta_1_t"],
                beta2=coefficients["beta_2_t"],
                epsilon=coefficients["epsilon"],
                grad=grad,
                use_locking=self._use_locking,
            )

        raise Exception("Optimizer is not Defined. Use adam or sgd.")

    def _resource_apply_sparse(self, grad, var, indices, apply_state=None):
        var_device, var_dtype = var.device, var.dtype.base_dtype
        coefficients = (apply_state or {}).get((var_device, var_dtype)) or self._fallback_apply_state(
            var_device, var_dtype
        )

        if self.currentOptimizer == CurrentOptimizer.SGD:
            momentum_var = self.get_slot(var, "momentum")
            return gen_training_ops.ResourceSparseApplyKerasMomentum(
                var=var.handle,
                accum=momentum_var.handle,
                lr=coefficients["lr_t"],
                grad=grad,
                indices=indices,
                momentum=coefficients["momentum"],
                use_locking=self._use_locking,
                use_nesterov=self.nesterov,
            )

        if self.currentOptimizer == CurrentOptimizer.ADAM:
            m = self.get_slot(var, "m")
            m_scaled_g_values = grad * coefficients["one_minus_beta_1_t"]
            m_t = state_ops.assign(m, m * coefficients["beta_1_t"], use_locking=self._use_locking)
            with ops.control_dependencies([m_t]):
                m_t = self._resource_scatter_add(m, indices, m_scaled_g_values)

            v = self.get_slot(var, "v")
            v_scaled_g_values = (grad * grad) * coefficients["one_minus_beta_2_t"]
            v_t = state_ops.assign(v, v * coefficients["beta_2_t"], use_locking=self._use_locking)
            with ops.control_dependencies([v_t]):
                v_t = self._resource_scatter_add(v, indices, v_scaled_g_values)

            if not self.amsgrad:
                v_sqrt = math_ops.sqrt(v_t)
                var_update = state_ops.assign_sub(
                    var, coefficients["lr"] * m_t / (v_sqrt + coefficients["epsilon"]), use_locking=self._use_locking
                )
                return control_flow_ops.group(*[var_update, m_t, v_t])

            v_hat = self.get_slot(var, "vhat")
            v_hat_t = math_ops.maximum(v_hat, v_t)
            with ops.control_dependencies([v_hat_t]):
                v_hat_t = state_ops.assign(v_hat, v_hat_t, use_locking=self._use_locking)
            v_hat_sqrt = math_ops.sqrt(v_hat_t)
            var_update = state_ops.assign_sub(
                var, coefficients["lr"] * m_t / (v_hat_sqrt + coefficients["epsilon"]), use_locking=self._use_locking
            )
            return control_flow_ops.group(*[var_update, m_t, v_t, v_hat_t])

        raise Exception("Optimizer is not Defined. Use adam or sgd.")

    def get_config(self) -> Dict[str, Any]:
        """Serialization of the object.

        Returns:
            Dictionary with the class' variable names as keys.
        """
        config = super(SwatsAdam, self).get_config()
        config.update(
            {
                "learning_rate": self._serialize_hyperparameter("learning_rate"),
                "decay": self._serialize_hyperparameter("decay"),
                "beta_1": self._serialize_hyperparameter("beta_1"),
                "beta_2": self._serialize_hyperparameter("beta_2"),
                "epsilon": self.epsilon,
                "amsgrad": self.amsgrad,
                "momentum": self._serialize_hyperparameter("momentum"),
                "nesterov": self.nesterov,
                "current_optimizer": self.currentOptimizer,
            }
        )
        return config


class SwatsNadam(optimizer_v2.OptimizerV2):
    """Initializer that can switch from NADAM to SGD and vice versa.

    SWATS_NADAM is inspired by the [SWATS](http://arxiv.org/abs/1712.07628) (switching from adam to sgd) initializer.
    During training, the optimizer can be changed by calling `switch_optimizer()` uppon the optimizer instance.

    This optimizer combines tensorflow's [NADAM](https://github.com/tensorflow/tensorflow/blob/85c8b2a817f95a3e979ecd1ed95bff1dc1335cff/tensorflow/python/keras/optimizer_v2/nadam.py)
    and [SGD](https://github.com/tensorflow/tensorflow/blob/85c8b2a817f95a3e979ecd1ed95bff1dc1335cff/tensorflow/python/keras/optimizer_v2/gradient_descent.py) implementations.

    **Example:**
    ```python
    import DeepSaki as ds
    optimizer = ds.optimizer.SWATS_NADAM(learning_rate=disc_learning_rate, beta_1=0.5, momentum=0.9, nesterov=False)
    optimizer.switch_optimizer()
    ```
    """

    _HAS_AGGREGATE_GRAD = True

    def __init__(
        self,
        learning_rate: float = 0.001,
        beta_1: float = 0.9,
        beta_2: float = 0.999,
        epsilon: float = 1e-7,
        momentum: float = 0.0,
        nesterov: bool = False,
        **kwargs: Any,
    ) -> None:
        """Initializes the `SwatsNadam` optimizer.

        Args:
            learning_rate (float, optional): Learning rate of the optimization step. Defaults to 0.001.
            beta_1 (float, optional): The exponential decay rate for the 1st moment estimates. Defaults to 0.9.
            beta_2 (float, optional): The exponential decay rate for the 2nd moment estimates. Defaults to 0.999.
            epsilon (float, optional): A small constant for numerical stability. Defaults to 1e-7.
            momentum (float, optional):  Hyperparameter >= 0 that accelerates gradient descent in the relevant direction
                and dampens oscillations. Vanilla gradient descent means no momentum. Defaults to 0.0.
            nesterov (bool, optional): Whether to apply Nesterov momentum. Defaults to False.
            kwargs: keyword arguments passed to the parent class.

        Raises:
            ValueError: if `momentum` is not provided in an expected form.
        """
        super(SwatsNadam, self).__init__(name="SwatsNadam", **kwargs)
        self._set_hyper("learning_rate", kwargs.get("lr", learning_rate))
        self._set_hyper("decay", self._initial_decay)
        self._set_hyper("beta_1", beta_1)
        self._set_hyper("beta_2", beta_2)
        self.epsilon = epsilon or backend_config.epsilon()
        self.currentOptimizer = CurrentOptimizer.NADAM

        self._momentum = False
        if isinstance(momentum, ops.Tensor) or callable(momentum) or momentum > 0:
            self._momentum = True
        if isinstance(momentum, (int, float)) and (momentum < 0 or momentum > 1):
            raise ValueError("`momentum` must be between [0, 1].")
        self._set_hyper("momentum", momentum)

        self.nesterov = nesterov
        self._m_cache = None

    def switch_optimizer(self) -> None:
        """Switches the current optimizer to either ADAM or SGD, depending on what was previously set."""
        if self.currentOptimizer == CurrentOptimizer.NADAM:
            self.currentOptimizer = CurrentOptimizer.SGD
        else:
            self.currentOptimizer = CurrentOptimizer.NADAM

    def _create_slots(self, var_list: List[tf.Variable]) -> None:
        # Create slots for the first and second moments.
        # Separate for-loops to respect the ordering of slot variables from v1.
        for var in var_list:
            self.add_slot(var, "m")
        for var in var_list:
            self.add_slot(var, "v")
        if self._m_cache is None:
            var_dtype = var_list[0].dtype.base_dtype
            self._m_cache = self.add_weight(
                "momentum_cache",
                shape=[],
                dtype=var_dtype,
                initializer="ones",
                trainable=False,
                aggregation=tf_variables.VariableAggregation.ONLY_FIRST_REPLICA,
            )
        self._weights.append(self._m_cache)

        if self._momentum:
            for var in var_list:
                self.add_slot(var, "momentum")

    def _prepare_local(self, var_device, var_dtype: tf.DType, apply_state) -> None:
        super(SwatsNadam, self)._prepare_local(var_device, var_dtype, apply_state)

        # From Nadam Optimizer!
        lr_t = array_ops.identity(self._get_hyper("learning_rate", var_dtype))
        beta_1_t = array_ops.identity(self._get_hyper("beta_1", var_dtype))
        beta_2_t = array_ops.identity(self._get_hyper("beta_2", var_dtype))
        local_step = math_ops.cast(self.iterations + 1, var_dtype)
        next_step = math_ops.cast(self.iterations + 2, var_dtype)
        decay_base = math_ops.cast(0.96, var_dtype)
        m_t = beta_1_t * (1.0 - 0.5 * (math_ops.pow(decay_base, self._initial_decay * local_step)))
        m_t_1 = beta_1_t * (1.0 - 0.5 * (math_ops.pow(decay_base, self._initial_decay * next_step)))
        m_schedule_new = math_ops.cast(self._m_cache_read, var_dtype) * m_t
        if var_dtype is self._m_cache.dtype:
            m_schedule_new = array_ops.identity(
                state_ops.assign(self._m_cache, m_schedule_new, use_locking=self._use_locking)
            )
        m_schedule_next = m_schedule_new * m_t_1

        apply_state[(var_device, var_dtype)] = {
            "lr_t": lr_t,
            "neg_lr_t": -lr_t,
            "epsilon": ops.convert_to_tensor_v2_with_dispatch(self.epsilon, var_dtype),
            "beta_1_t": beta_1_t,
            "beta_2_t": beta_2_t,
            "m_t": m_t,
            "m_t_1": m_t_1,
            "one_minus_beta_1_t": 1 - beta_1_t,
            "one_minus_beta_2_t": 1 - beta_2_t,
            "one_minus_m_t": 1.0 - m_t,
            "one_minus_m_schedule_new": 1.0 - m_schedule_new,
            "one_minus_m_schedule_next": 1.0 - m_schedule_next,
            "v_t_prime_denominator": 1.0 - math_ops.pow(beta_2_t, local_step),
        }

        # from SGD Optimizer
        apply_state[(var_device, var_dtype)]["momentum"] = array_ops.identity(self._get_hyper("momentum", var_dtype))

    def _prepare(self, var_list: List[tf.Variable]):
        # Get the value of the momentum cache before starting to apply gradients.
        self._m_cache_read = array_ops.identity(self._m_cache)
        return super(SwatsNadam, self)._prepare(var_list)

    def _resource_apply_dense(self, grad, var, apply_state=None):
        var_device, var_dtype = var.device, var.dtype.base_dtype
        coefficients = (apply_state or {}).get((var_device, var_dtype)) or self._fallback_apply_state(
            var_device, var_dtype
        )

        # SGD Optimizer!
        if self.currentOptimizer == CurrentOptimizer.SGD:
            if self._momentum:
                momentum_var = self.get_slot(var, "momentum")
                return gen_training_ops.ResourceApplyKerasMomentum(
                    var=var.handle,
                    accum=momentum_var.handle,
                    lr=coefficients["lr_t"],
                    grad=grad,
                    momentum=coefficients["momentum"],
                    use_locking=self._use_locking,
                    use_nesterov=self.nesterov,
                )
            return gen_training_ops.ResourceApplyGradientDescent(
                var=var.handle, alpha=coefficients["lr_t"], delta=grad, use_locking=self._use_locking
            )
        # Nadam Optimizer!
        if self.currentOptimizer == CurrentOptimizer.NADAM:
            m = self.get_slot(var, "m")
            v = self.get_slot(var, "v")
            g_prime = grad / coefficients["one_minus_m_schedule_new"]
            m_t = coefficients["beta_1_t"] * m + coefficients["one_minus_beta_1_t"] * grad
            m_t = state_ops.assign(m, m_t, use_locking=self._use_locking)
            m_t_prime = m_t / coefficients["one_minus_m_schedule_next"]
            v_t = coefficients["beta_2_t"] * v + coefficients["one_minus_beta_2_t"] * math_ops.square(grad)
            v_t = state_ops.assign(v, v_t, use_locking=self._use_locking)
            v_t_prime = v_t / coefficients["v_t_prime_denominator"]
            m_t_bar = coefficients["one_minus_m_t"] * g_prime + coefficients["m_t_1"] * m_t_prime
            var_t = var - coefficients["lr_t"] * m_t_bar / (math_ops.sqrt(v_t_prime) + coefficients["epsilon"])
            return state_ops.assign(var, var_t, use_locking=self._use_locking).op

        raise Exception("Optimizer is not Defined. Use nadam or sgd.")

    def _resource_apply_sparse(self, grad, var, indices, apply_state=None):
        var_device, var_dtype = var.device, var.dtype.base_dtype
        coefficients = (apply_state or {}).get((var_device, var_dtype)) or self._fallback_apply_state(
            var_device, var_dtype
        )

        # SGD Part
        if self.currentOptimizer == CurrentOptimizer.SGD:
            momentum_var = self.get_slot(var, "momentum")
            return gen_training_ops.ResourceSparseApplyKerasMomentum(
                var=var.handle,
                accum=momentum_var.handle,
                lr=coefficients["lr_t"],
                grad=grad,
                indices=indices,
                momentum=coefficients["momentum"],
                use_locking=self._use_locking,
                use_nesterov=self.nesterov,
            )
        # Nadam Part
        if self.currentOptimizer == CurrentOptimizer.NADAM:
            m = self.get_slot(var, "m")
            v = self.get_slot(var, "v")

            g_prime = grad / coefficients["one_minus_m_schedule_new"]

            m_scaled_g_values = grad * coefficients["one_minus_beta_1_t"]
            m_t = state_ops.assign(m, m * coefficients["beta_1_t"], use_locking=self._use_locking)

            with ops.control_dependencies([m_t]):
                m_t = self._resource_scatter_add(m, indices, m_scaled_g_values)
                m_t_slice = array_ops.gather(m_t, indices)

            m_t_prime = m_t_slice / coefficients["one_minus_m_schedule_next"]
            m_t_bar = coefficients["one_minus_m_t"] * g_prime + coefficients["m_t_1"] * m_t_prime

            v_scaled_g_values = (grad * grad) * coefficients["one_minus_beta_2_t"]
            v_t = state_ops.assign(v, v * coefficients["beta_2_t"], use_locking=self._use_locking)

            with ops.control_dependencies([v_t]):
                v_t = self._resource_scatter_add(v, indices, v_scaled_g_values)
                v_t_slice = array_ops.gather(v_t, indices)

            v_t_prime = v_t_slice / coefficients["v_t_prime_denominator"]
            v_prime_sqrt_plus_eps = math_ops.sqrt(v_t_prime) + coefficients["epsilon"]

            var_update = self._resource_scatter_add(
                var, indices, coefficients["neg_lr_t"] * m_t_bar / v_prime_sqrt_plus_eps
            )
            return control_flow_ops.group(*[var_update, m_t_bar, v_t])

        raise Exception("Optimizer is not Defined. Use nadam or sgd.")

    def get_config(self) -> Dict[str, Any]:
        """Serialization of the object.

        Returns:
            Dictionary with the class' variable names as keys.
        """
        config = super(SwatsNadam, self).get_config()
        config.update(
            {
                "learning_rate": self._serialize_hyperparameter("learning_rate"),
                "decay": self._serialize_hyperparameter("decay"),
                "beta_1": self._serialize_hyperparameter("beta_1"),
                "beta_2": self._serialize_hyperparameter("beta_2"),
                "epsilon": self.epsilon,
                "momentum": self._serialize_hyperparameter("momentum"),
                "nesterov": self.nesterov,
                "current_optimizer": self.currentOptimizer,
            }
        )
        return config



from abc import ABC
from abc import abstractmethod
from typing import Callable

import numpy as np
import tensorflow as tf

from DeepSaki.types.losses_enums import LossCalcType
from DeepSaki.types.losses_enums import LossType

class ImageBasedLoss(tf.keras.losses.Loss, ABC):
    """Abstract base class for image based losses.

    Sub-classes must override `_error_func(self, tensor1: tf.Tensor, tensor2: tf.Tensor) -> tf.Tensor`, the actuall
    loss function that callculates the loss between two tensors.
    """

    def __init__(
        self,
        global_batch_size: int,
        calculation_type: LossCalcType = LossCalcType.PER_IMAGE,
        normalize_depth_channel: bool = False,
        loss_reduction: tf.keras.losses.Reduction = tf.keras.losses.Reduction.AUTO,
    ) -> None:
        """Initializes the abstract base class `ImageBasedLoss`.

        Args:
            global_batch_size (int): Batch size considering all workers running in parallel in a data parallel setup
            calculation_type (LossCalcType, optional): Determines how the loss is calculated.
                Defaults to LossCalcType.PER_IMAGE.
            normalize_depth_channel (bool, optional): For RGBD images, the weight of depth is increased by multiplying
                the depth by the number of color channels. Defaults to False.
            loss_reduction (tf.keras.losses.Reduction, optional): Determines how the loss is reduced. Defaults to
                tf.keras.losses.Reduction.AUTO.

        Raises:
            ValueError: if `calculation_type` is not a valid option.
        """
        super(ImageBasedLoss, self).__init__(reduction=loss_reduction)

        match calculation_type:
            case LossCalcType.PER_CHANNEL:
                self.loss_calc_func = self._calc_loss_per_channel
            case LossCalcType.PER_IMAGE:
                self.loss_calc_func = self._calc_loss_per_image
            case _:
                raise ValueError(f"Pixel distance type '{calculation_type}' is not supported.")

        self.global_batch_size = global_batch_size
        self.normalize_depth_channel = normalize_depth_channel

    @abstractmethod
    def _error_func(self, tensor1: tf.Tensor, tensor2: tf.Tensor) -> tf.Tensor:
        pass

    def _get_channel_weights(self, img_shape: tf.TensorShape, normalize_last_channel: bool) -> np.ndarray:
        channel_weight = np.ones(img_shape[-1])
        if normalize_last_channel:
            # set weight of the depth channel according to the number of color channels: e.g. for RGB = 3
            channel_weight[-1] = len(channel_weight) - 1
        return channel_weight

    def _calc_loss_per_channel(
        self,
        img1: tf.Tensor,
        img2: tf.Tensor,
        error_func: Callable[[tf.Tensor, tf.Tensor], tf.Tensor],
    ) -> tf.Tensor:
        # initialize all weights with 1
        channel_weight = self._get_channel_weights(img1.shape, self.normalize_depth_channel)
        loss = 0.0
        for i in range(img1.shape[-1]):
            loss += channel_weight[i] * error_func(img1[:, :, :, i], img2[:, :, :, i]) * (1.0 / self.global_batch_size)
        return loss

    def _calc_loss_per_image(
        self,
        img1: tf.Tensor,
        img2: tf.Tensor,
        error_func: Callable[[tf.Tensor, tf.Tensor], tf.Tensor],
    ) -> tf.Tensor:
        return error_func(img1, img2) * (1.0 / self.global_batch_size)

    def call(self, img1: tf.Tensor, img2: tf.Tensor) -> tf.Tensor:
        """Calculates the image loss between `img1` and `img2` using the overriden `error_func` from the sub-class.

        Args:
            img1 (tf.Tensor): Image of shape (`batch_size`, `height`, `width`,`channel`).
            img2 (tf.Tensor): Image of shape (`batch_size`, `height`, `width`,`channel`).

        Returns:
            Tensor containing the loss.
        """
        img1 = tf.cast(img1, tf.dtypes.float32)
        img2 = tf.cast(img2, tf.dtypes.float32)

        return self.loss_calc_func(img1, img2, self._error_func)


class PixelDistanceLoss(ImageBasedLoss):
    """Calculates a pixel distance loss (per pixel loss) of two images of the shape (batch, height, width, channels).

    The distance is either a mean squared error (MSE) or a mean absolute error (MAE).
    """

    def __init__(
        self,
        global_batch_size: int,
        calculation_type: LossCalcType = LossCalcType.PER_IMAGE,
        normalize_depth_channel: bool = False,
        loss_type: LossType = LossType.MAE,
        loss_reduction: tf.keras.losses.Reduction = tf.keras.losses.Reduction.AUTO,
    ) -> None:
        """Initializes the `PixelDistanceLoss`.

        Args:
            global_batch_size (int): Batch size considering all workers running in parallel in a data parallel setup
            calculation_type (LossCalcType, optional): Determines how the loss is calculated.
                Defaults to LossCalcType.PER_IMAGE.
            normalize_depth_channel (bool, optional): For RGBD images, the weight of depth is increased by multiplying
                the depth by the number of color channels. Defaults to False.
            loss_type (LossType, optional): Loss to apply. Defaults to LossType.MAE.
            loss_reduction (tf.keras.losses.Reduction, optional): Determines how the loss is reduced. Defaults to
                tf.keras.losses.Reduction.AUTO.

        Raises:
            ValueError: if `loss_type` is not a valid option.
        """
        match loss_type:
            case LossType.MAE:
                self.error_func_type = tf.abs
            case LossType.MSE:
                self.error_func_type = tf.square
            case _:
                raise ValueError(f"Parameter loss_type={loss_type} is not supported.")

        super(PixelDistanceLoss, self).__init__(
            global_batch_size=global_batch_size,
            calculation_type=calculation_type,
            normalize_depth_channel=normalize_depth_channel,
            loss_reduction=loss_reduction,
        )

    # Override abstract method
    def _error_func(self, tensor1: tf.Tensor, tensor2: tf.Tensor) -> tf.Tensor:
        return tf.reduce_mean(self.error_func_type(tensor1 - tensor2))


class StructuralSimilarityLoss(ImageBasedLoss):
    r"""Calculates the structural similarity (SSIM) loss of two images of the shape (batch, height, width, channels).

    The structural similarity loss between two images $x$ and $y$ is defined as: $\mathcal{L}_{SSIM}=1-SSIM(x,y)$.

    SSIM compares contrast, luminance, and structure of two images using statistical parameters i.e., the mean $\mu$,
    the variance $\sigma$, and the covariance $\sigma_{x,y}$ of both images. The SSIM of two images $x$ and $y$ is
    defined as:

    $$
    SSIM(x,y)=
    \underbrace{\left[ \frac{2\mu_x\mu_y+C_1}{\mu_x^2+\mu_y^2 +C_1}\right]^{\alpha}}_\text{contrast} \cdot
    \underbrace{\left[ \frac{2\sigma_x\sigma_y+C_2}{\sigma_x^2+\sigma_y^2 +C_2}\right]^{\beta}}_\text{luminance} \cdot
    \underbrace{\left[ \frac{\sigma_{x,y}+C_3}{\sigma_x\sigma_y+C_3}\right]^{\gamma}}_\text{structure},
    $$

    where $\alpha$, $\beta$ and $\gamma$ are hyperparameters to give relative importance to individual terms and $C_1$,
    $C_2$ and $C_3$ are constants that must be chosen.
    """

    def __init__(
        self,
        global_batch_size: int,
        calculation_type: LossCalcType = LossCalcType.PER_IMAGE,
        normalize_depth_channel: bool = False,
        alpha: float = 1.0,
        beta: float = 1.0,
        gamma: float = 1.0,
        c1: float = 0.0001,
        c2: float = 0.0009,
        loss_reduction: tf.keras.losses.Reduction = tf.keras.losses.Reduction.AUTO,
    ) -> None:
        """Dunder method to initialize `StructuralSimilarityLoss`.

        Args:
            global_batch_size (int): Batch size considering all workers running in parallel in a data parallel setup.
            calculation_type (LossCalcType, optional): Determines how the loss is calculated.
                Defaults to LossCalcType.PER_IMAGE.
            normalize_depth_channel (bool, optional): For RGBD images, the weight of depth is increased by multiplying
                the depth by the number of color channels. Defaults to False.
            alpha (float, optional): Weighting factor for contrast. Defaults to 1.0.
            beta (float, optional): Weighting factor for luminance. Defaults to 1.0.
            gamma (float, optional): Weighting factor for structure. Defaults to 1.0.
            c1 (float, optional): Constant considered in contrast calculation. Defaults to 0.0001.
            c2 (float, optional): Constant considered in luminance calculation. Defaults to 0.0009.
            loss_reduction (tf.keras.losses.Reduction, optional): Determines how the loss is reduced. Defaults to
                tf.keras.losses.Reduction.AUTO.
        """
        super(StructuralSimilarityLoss, self).__init__(
            global_batch_size=global_batch_size,
            calculation_type=calculation_type,
            normalize_depth_channel=normalize_depth_channel,
            loss_reduction=loss_reduction,
        )

        self.alpha = alpha
        self.beta = beta
        self.gamma = gamma
        self.c1 = c1
        self.c2 = c2
        self.c3 = c2 / 2

    def _calc_ssim_loss(self, tensor1: tf.Tensor, tensor2: tf.Tensor) -> tf.Tensor:
        """Calculates structural similarity loss between `tensor1` and `tensor2`.

        Args:
            tensor1 (tf.Tensor): Image of shape (`batch_size`, `height`, `width`,`channel`).
            tensor2 (tf.Tensor): Image of shape (`batch_size`, `height`, `width`,`channel`).

        Returns:
            Structural similarity loss between `tensor1` and `tensor2`.
        """
        mu1 = tf.reduce_mean(tensor1)  # mean
        mu2 = tf.reduce_mean(tensor2)
        sigma1 = tf.reduce_mean((tensor1 - mu1) ** 2) ** 0.5  # standard deviation
        sigma2 = tf.reduce_mean((tensor2 - mu2) ** 2) ** 0.5
        covar = tf.reduce_mean((tensor1 - mu1) * (tensor2 - mu2))  # covariance

        luminance = (2 * mu1 * mu2 + self.c1) / (mu1**2 + mu2**2 + self.c1)
        contrast = (2 * sigma1 * sigma2 + self.c2) / (sigma1**2 + sigma2**2 + self.c2)
        structure = (covar + self.c3) / (sigma1 * sigma2 + self.c3)

        ssim = luminance**self.alpha * contrast**self.beta * structure**self.gamma
        return 1 - ssim

    # Override abstract method
    def _error_func(self, tensor1: tf.Tensor, tensor2: tf.Tensor) -> tf.Tensor:
        return self._calc_ssim_loss(tensor1, tensor2)



import os
from contextlib import nullcontext as does_not_raise

import pytest

os.environ["TF_CPP_MIN_LOG_LEVEL"] = "2"  # deactivate tensorflow warnings and infos. Keep Errors
import tensorflow as tf

from DeepSaki.layers.layer_composites import Conv2DBlock
from DeepSaki.layers.layer_composites import Conv2DSplitted
from DeepSaki.layers.layer_composites import DenseBlock
from DeepSaki.layers.layer_composites import DownSampleBlock
from DeepSaki.types.layers_enums import PaddingType, UpSampleType, DownSampleType
from DeepSaki.layers.layer_composites import ResBlockDown
from DeepSaki.layers.layer_composites import ResBlockUp
from DeepSaki.layers.layer_composites import ResidualBlock
from DeepSaki.layers.layer_composites import ScalarGatedSelfAttention
from DeepSaki.layers.layer_composites import ScaleLayer
from DeepSaki.layers.layer_composites import UpSampleBlock
from tests.DeepSaki_test.layers_test.layers_test import CommonLayerChecks
from tests.DeepSaki_test.layers_test.layers_test import DeepSakiLayerChecks
from tests.DeepSaki_test.layers_test.mocked_layers import _mock_scale_layer

# TODO: ensure I only test inputs relevant for that particular layer and that are relevant for the test. esp. when testing the output shape.
# TODO: change expected output shape to be an equation. In that way it is easier to understand later and it is always the same shape, which is important for layers using this layer.


class TestConv2DSplitted(DeepSakiLayerChecks):
    @pytest.fixture()
    def conv_2d_splitted(self):
        return Conv2DSplitted()

    @pytest.mark.parametrize("use_spec_norm", [True, False])
    @pytest.mark.parametrize("use_bias", [True, False])
    @pytest.mark.parametrize(
        ("input", "filters", "kernels", "strides", "padding", "expected_shape"),
        [
            (tf.TensorShape((1, 16, 16, 3)), 3, 3, (1, 1), "same", tf.TensorShape((1, 16, 16, 3))),
            (tf.TensorShape((8, 16, 16, 3)), 5, 5, (1, 1), "same", tf.TensorShape((8, 16, 16, 5))),
            (tf.TensorShape((1, 16, 16, 3)), 3, 3, (2, 2), "same", tf.TensorShape((1, 8, 8, 3))),
            (tf.TensorShape((1, 16, 16, 3)), 3, 5, (1, 2), "same", tf.TensorShape((1, 16, 8, 3))),
            (tf.TensorShape((1, 16, 16, 3)), 3, 3, (2, 1), "same", tf.TensorShape((1, 8, 16, 3))),
            (tf.TensorShape((1, 16, 16, 3)), 3, 5, (1, 1), "valid", tf.TensorShape((1, 12, 12, 3))),
            (tf.TensorShape((1, 16, 16, 3)), 5, 3, (1, 1), "valid", tf.TensorShape((1, 14, 14, 5))),
            (tf.TensorShape((1, 16, 16, 3)), 3, 3, (2, 2), "valid", tf.TensorShape((1, 7, 7, 3))),
        ],
    )
    def test_call_correct_output_shape(
        self, input, filters, strides, padding, use_bias, use_spec_norm, kernels, expected_shape
    ):
        layer_instance = Conv2DSplitted(filters, kernels, use_spec_norm, strides, use_bias, padding)
        CommonLayerChecks.has_call_correct_output_shape(layer_instance, input, expected_shape)

    @pytest.mark.parametrize(
        ("input_shape", "expected_context"),
        [
            (tf.TensorShape((8, 64, 64, 4)), does_not_raise()),
            (tf.TensorShape((1, 32, 32, 3)), does_not_raise()),
            (tf.TensorShape((8, 64, 64)), pytest.raises(ValueError)),
            (tf.TensorShape((8, 64)), pytest.raises(ValueError)),
            (tf.TensorShape((8)), pytest.raises(ValueError)),
            (tf.TensorShape((8, 64, 64, 4, 5)), pytest.raises(ValueError)),
            (tf.TensorShape((8, 64, 64, 4, 5, 6)), pytest.raises(ValueError)),
        ],
    )
    def test_call_raises_error_wrong_input_spec(self, conv_2d_splitted, input_shape, expected_context):
        CommonLayerChecks.does_call_raises_error_wrong_input_spec(conv_2d_splitted, input_shape, expected_context)


class TestConv2DBlock(DeepSakiLayerChecks):
    @pytest.fixture()
    def conv_2d_block(self):
        return Conv2DBlock()

    @pytest.mark.parametrize("use_spec_norm", [True, False])
    @pytest.mark.parametrize("split_kernels", [True, False])
    @pytest.mark.parametrize("use_bias", [False])
    @pytest.mark.parametrize("final_activation", [True, False])
    @pytest.mark.parametrize("apply_final_normalization", [True, False])
    @pytest.mark.parametrize("number_of_blocks", [1, 2])
    @pytest.mark.parametrize(
        ("input_shape", "filters", "kernels", "strides"),
        [
            (tf.TensorShape((1, 16, 16, 3)), 3, 3, (1, 1)),
            (tf.TensorShape((8, 16, 16, 3)), 5, 5, (1, 1)),
            (tf.TensorShape((1, 16, 16, 3)), 3, 3, (2, 2)),
            (tf.TensorShape((8, 16, 16, 3)), 5, 5, (1, 2)),
            (tf.TensorShape((1, 16, 16, 3)), 1, 3, (2, 1)),
        ],
    )
    def test_call_correct_output_shape(
        self,
        input_shape,
        filters,
        strides,
        use_bias,
        use_spec_norm,
        kernels,
        number_of_blocks,
        split_kernels,
        final_activation,
        apply_final_normalization,
    ):
        layer_instance = Conv2DBlock(
            filters,
            kernels,
            split_kernels,
            number_of_blocks,
            final_activation=final_activation,
            apply_final_normalization=apply_final_normalization,
            use_spec_norm=use_spec_norm,
            strides=strides,
            use_bias=use_bias,
        )
        expected_shape = (
            input_shape[0],
            input_shape[1] // (strides[0] ** number_of_blocks),
            input_shape[2] // (strides[1] ** number_of_blocks),
            filters,
        )
        CommonLayerChecks.has_call_correct_output_shape(layer_instance, input_shape, expected_shape)

    @pytest.mark.parametrize(
        ("input_shape", "expected_context"),
        [
            (tf.TensorShape((8, 64, 64, 4)), does_not_raise()),
            (tf.TensorShape((1, 32, 32, 3)), does_not_raise()),
            (tf.TensorShape((8, 64, 64)), pytest.raises(ValueError)),
            (tf.TensorShape((8, 64)), pytest.raises(ValueError)),
            (tf.TensorShape((8)), pytest.raises(ValueError)),
            (tf.TensorShape((8, 64, 64, 4, 5)), pytest.raises(ValueError)),
            (tf.TensorShape((8, 64, 64, 4, 5, 6)), pytest.raises(ValueError)),
        ],
    )
    def test_call_raises_error_wrong_input_spec(self, conv_2d_block, input_shape, expected_context):
        CommonLayerChecks.does_call_raises_error_wrong_input_spec(conv_2d_block, input_shape, expected_context)

    @pytest.mark.parametrize("number_of_blocks", [1, 2, 3])
    def test_number_of_blocks_correct(self, number_of_blocks):
        layer_instance = Conv2DBlock(number_of_blocks=number_of_blocks)
        assert len(layer_instance.blocks) == number_of_blocks

    @pytest.mark.parametrize("number_of_blocks", [1, 2, 3])
    @pytest.mark.parametrize(
        ("apply_final_normalization", "final_activation", "expected_final_len"),
        [
            (True, True, 4),
            (True, False, 3),
            (False, True, 3),
            (False, False, 2),
        ],
    )
    def test_final_normalization_activation_correctly_set(
        self, number_of_blocks, apply_final_normalization, final_activation, expected_final_len
    ):
        layer_instance = Conv2DBlock(
            number_of_blocks=number_of_blocks,
            apply_final_normalization=apply_final_normalization,
            final_activation=final_activation,
        )
        assert len(layer_instance.blocks[-1]) == expected_final_len


class TestDenseBlock(DeepSakiLayerChecks):
    @pytest.fixture()
    def dense_block(self):
        return DenseBlock(16)

    @pytest.mark.parametrize("use_spec_norm", [True, False])
    @pytest.mark.parametrize("use_bias", [False])
    @pytest.mark.parametrize("final_activation", [True, False])
    @pytest.mark.parametrize("apply_final_normalization", [True, False])
    @pytest.mark.parametrize("number_of_blocks", [1, 2])
    @pytest.mark.parametrize("units", [8, 16, 32])
    @pytest.mark.parametrize(
        ("input_shape"),
        [
            (tf.TensorShape((1, 16, 16, 3))),
            (tf.TensorShape((8, 32, 20))),
            (tf.TensorShape((8, 32))),
        ],
    )
    def test_call_correct_output_shape(
        self, input_shape, units, use_bias, use_spec_norm, number_of_blocks, final_activation, apply_final_normalization
    ):
        layer_instance = DenseBlock(
            units,
            number_of_blocks,
            final_activation=final_activation,
            apply_final_normalization=apply_final_normalization,
            use_spec_norm=use_spec_norm,
            use_bias=use_bias,
        )
        expected_shape = [*input_shape[0:-1], units]
        CommonLayerChecks.has_call_correct_output_shape(layer_instance, input_shape, expected_shape)

    @pytest.mark.parametrize(
        ("input_shape", "expected_context"),
        [
            (tf.TensorShape((8, 64, 64, 4)), does_not_raise()),
            (tf.TensorShape((1, 32, 32, 3)), does_not_raise()),
            (tf.TensorShape((8, 64, 64)), does_not_raise()),
            (tf.TensorShape((8, 64)), does_not_raise()),
            (tf.TensorShape((8)), pytest.raises(ValueError)),
            (tf.TensorShape((8, 64, 64, 4, 5)), does_not_raise()),
            (tf.TensorShape((8, 64, 64, 4, 5, 6)), does_not_raise()),
        ],
    )
    def test_call_raises_error_wrong_input_spec(self, dense_block, input_shape, expected_context):
        CommonLayerChecks.does_call_raises_error_wrong_input_spec(dense_block, input_shape, expected_context)

    @pytest.mark.parametrize("number_of_blocks", [1, 2, 3])
    def test_number_of_blocks_correct(self, number_of_blocks):
        layer_instance = DenseBlock(units=16, number_of_blocks=number_of_blocks)
        assert len(layer_instance.blocks) == number_of_blocks

    @pytest.mark.parametrize("number_of_blocks", [1, 2, 3])
    @pytest.mark.parametrize(
        ("apply_final_normalization", "final_activation", "expected_final_len"),
        [
            (True, True, 3),
            (True, False, 2),
            (False, True, 2),
            (False, False, 1),
        ],
    )
    def test_final_normalization_activation_correctly_set(
        self, number_of_blocks, apply_final_normalization, final_activation, expected_final_len
    ):
        layer_instance = DenseBlock(
            units=16,
            number_of_blocks=number_of_blocks,
            apply_final_normalization=apply_final_normalization,
            final_activation=final_activation,
        )
        assert len(layer_instance.blocks[-1]) == expected_final_len


class TestDownSamplingBlock(DeepSakiLayerChecks):
    @pytest.fixture()
    def downsample_block(self):
        return DownSampleBlock()

    @pytest.mark.parametrize(
        "downsampling",
        [
            DownSampleType.AVG_POOLING,
            DownSampleType.MAX_POOLING,
            DownSampleType.CONV_STRIDE_2,
            DownSampleType.SPACE_TO_DEPTH,
        ],
    )
    @pytest.mark.parametrize("kernels", [3, 5])
    @pytest.mark.parametrize("use_bias", [True, False])
    @pytest.mark.parametrize(
        ("input_shape"),
        [
            tf.TensorShape((1, 16, 16, 4)),
        ],
    )
    def test_call_correct_output_shape(self, input_shape, downsampling, kernels, use_bias):
        layer_instance = DownSampleBlock(downsampling=downsampling, kernels=kernels, use_bias=use_bias)
        expected_shape = [
            input_shape[0],
            input_shape[1] // 2,
            input_shape[2] // 2,
            input_shape[3],
        ]
        CommonLayerChecks.has_call_correct_output_shape(layer_instance, input_shape, expected_shape)

    @pytest.mark.parametrize(
        ("input_shape", "expected_context"),
        [
            (tf.TensorShape((8, 64, 64, 4)), does_not_raise()),
            (tf.TensorShape((1, 32, 32, 3)), does_not_raise()),
            (tf.TensorShape((8, 64, 64)), pytest.raises(ValueError)),
            (tf.TensorShape((8, 64)), pytest.raises(ValueError)),
            (tf.TensorShape((8)), pytest.raises(ValueError)),
            (tf.TensorShape((8, 64, 64, 4, 5)), pytest.raises(ValueError)),
            (tf.TensorShape((8, 64, 64, 4, 5, 6)), pytest.raises(ValueError)),
        ],
    )
    def test_call_raises_error_wrong_input_spec(self, downsample_block, input_shape, expected_context):
        CommonLayerChecks.does_call_raises_error_wrong_input_spec(downsample_block, input_shape, expected_context)

    @pytest.mark.parametrize(
        ("downsampling", "expected_context"),
        [
            (DownSampleType.CONV_STRIDE_2, does_not_raise()),
            (DownSampleType.MAX_POOLING, does_not_raise()),
            (DownSampleType.AVG_POOLING, does_not_raise()),
            (DownSampleType.SPACE_TO_DEPTH, does_not_raise()),
            ("Any other String", pytest.raises(ValueError)),
        ],
    )
    def test_build_raises_error_wrong_downsampling(self, downsampling, expected_context):
        layer_instance = DownSampleBlock(downsampling=downsampling)
        with expected_context:
            layer_instance.build(input_shape=(1, 16, 16, 4))


class TestUpSamplingBlock(DeepSakiLayerChecks):
    @pytest.fixture()
    def upsample_block(self):
        return UpSampleBlock()

    @pytest.mark.parametrize(
        "upsampling", [UpSampleType.RESAMPLE_AND_CONV, UpSampleType.TRANSPOSE_CONV, UpSampleType.DEPTH_TO_SPACE]
    )
    @pytest.mark.parametrize("kernels", [3, 5])
    @pytest.mark.parametrize("use_bias", [True, False])
    @pytest.mark.parametrize(
        ("input_shape"),
        [
            tf.TensorShape((1, 16, 16, 4)),
        ],
    )
    def test_call_correct_output_shape(self, input_shape, upsampling, kernels, use_bias):
        layer_instance = UpSampleBlock(upsampling=upsampling, kernels=kernels, use_bias=use_bias)
        expected_shape = [
            input_shape[0],
            input_shape[1] * 2,
            input_shape[2] * 2,
            input_shape[3],
        ]
        CommonLayerChecks.has_call_correct_output_shape(layer_instance, input_shape, expected_shape)

    @pytest.mark.parametrize(
        ("input_shape", "expected_context"),
        [
            (tf.TensorShape((8, 64, 64, 4)), does_not_raise()),
            (tf.TensorShape((1, 32, 32, 3)), does_not_raise()),
            (tf.TensorShape((8, 64, 64)), pytest.raises(ValueError)),
            (tf.TensorShape((8, 64)), pytest.raises(ValueError)),
            (tf.TensorShape((8)), pytest.raises(ValueError)),
            (tf.TensorShape((8, 64, 64, 4, 5)), pytest.raises(ValueError)),
            (tf.TensorShape((8, 64, 64, 4, 5, 6)), pytest.raises(ValueError)),
        ],
    )
    def test_call_raises_error_wrong_input_spec(self, upsample_block, input_shape, expected_context):
        CommonLayerChecks.does_call_raises_error_wrong_input_spec(upsample_block, input_shape, expected_context)

    @pytest.mark.parametrize(
        ("upsampling", "expected_context"),
        [
            (UpSampleType.RESAMPLE_AND_CONV, does_not_raise()),
            (UpSampleType.TRANSPOSE_CONV, does_not_raise()),
            (UpSampleType.DEPTH_TO_SPACE, does_not_raise()),
            ("Any other String", pytest.raises(ValueError)),
        ],
    )
    def test_build_raises_error_wrong_downsampling(self, upsampling, expected_context):
        layer_instance = UpSampleBlock(upsampling=upsampling)
        with expected_context:
            layer_instance.build(input_shape=(1, 16, 16, 4))


class TestResidualBlock(DeepSakiLayerChecks):
    @pytest.fixture()
    def residual_block(self):
        return ResidualBlock()

    @pytest.mark.parametrize(
        ("input_shape", "expected_context"),
        [
            (tf.TensorShape((8, 64, 64, 4)), does_not_raise()),
            (tf.TensorShape((1, 32, 32, 3)), does_not_raise()),
            (tf.TensorShape((8, 64, 64)), pytest.raises(ValueError)),
            (tf.TensorShape((8, 64)), pytest.raises(ValueError)),
            (tf.TensorShape((8)), pytest.raises(ValueError)),
            (tf.TensorShape((8, 64, 64, 4, 5)), pytest.raises(ValueError)),
            (tf.TensorShape((8, 64, 64, 4, 5, 6)), pytest.raises(ValueError)),
        ],
    )
    def test_call_raises_error_wrong_input_spec(self, residual_block, input_shape, expected_context):
        CommonLayerChecks.does_call_raises_error_wrong_input_spec(residual_block, input_shape, expected_context)

    @pytest.mark.parametrize("use_spec_norm", [False])
    @pytest.mark.parametrize("use_bias", [True])
    @pytest.mark.parametrize("activation", ["leaky_relu"])
    @pytest.mark.parametrize("kernels", [3])
    @pytest.mark.parametrize("dropout_rate", [0.0, 0.5])
    @pytest.mark.parametrize("number_of_blocks", [1, 2])
    @pytest.mark.parametrize("residual_cardinality", [1, 3])
    @pytest.mark.parametrize("padding", [PaddingType.ZERO])
    @pytest.mark.parametrize(
        ("input", "filters", "expected_shape"),
        [
            (tf.TensorShape((1, 16, 16, 3)), 3, tf.TensorShape((1, 16, 16, 3))),
            (tf.TensorShape((1, 16, 16, 3)), 8, tf.TensorShape((1, 16, 16, 8))),
            (tf.TensorShape((1, 16, 16, 3)), 5, tf.TensorShape((1, 16, 16, 5))),
            (tf.TensorShape((4, 8, 8, 4)), 12, tf.TensorShape((4, 8, 8, 12))),
        ],
    )
    def test_call_correct_output_shape(
        self,
        input,
        expected_shape,
        filters,
        kernels,
        activation,
        number_of_blocks,
        use_spec_norm,
        residual_cardinality,
        dropout_rate,
        use_bias,
        padding,
    ):
        layer_instance = ResidualBlock(
            filters,
            kernels,
            activation,
            number_of_blocks,
            use_spec_norm,
            residual_cardinality,
            dropout_rate,
            use_bias,
            padding,
        )
        CommonLayerChecks.has_call_correct_output_shape(layer_instance, input, expected_shape)

    @pytest.mark.parametrize("number_of_blocks", [1, 2, 3])
    def test_number_of_blocks_correct(self, number_of_blocks):
        layer = ResidualBlock(number_of_blocks=number_of_blocks)
        assert len(layer.blocks) == number_of_blocks

    @pytest.mark.parametrize("residual_cardinality", [1, 2, 3])
    @pytest.mark.parametrize("number_of_blocks", [1, 2, 3])
    def test_residual_cardinality_correct(self, number_of_blocks, residual_cardinality):
        layer = ResidualBlock(number_of_blocks=number_of_blocks, residual_cardinality=residual_cardinality)
        for block in layer.blocks:
            assert len(block) == residual_cardinality


class TestResBlockDown(DeepSakiLayerChecks):
    @pytest.fixture()
    def resblock_down(self):
        return ResBlockDown()

    @pytest.mark.parametrize(
        ("input_shape"),
        [
            tf.TensorShape((1, 16, 16, 4)),
            tf.TensorShape((8, 32, 32, 3)),
            tf.TensorShape((5, 32, 16, 12)),
            tf.TensorShape((3, 16, 32, 8)),
        ],
    )
    def test_call_correct_output_shape(self, input_shape):
        layer_instance = ResBlockDown()
        expected_shape = [
            input_shape[0],
            input_shape[1] // 2,
            input_shape[2] // 2,
            input_shape[3],
        ]
        CommonLayerChecks.has_call_correct_output_shape(layer_instance, input_shape, expected_shape)

    @pytest.mark.parametrize(
        ("input_shape", "expected_context"),
        [
            (tf.TensorShape((8, 64, 64, 4)), does_not_raise()),
            (tf.TensorShape((1, 32, 32, 3)), does_not_raise()),
            (tf.TensorShape((8, 64, 64)), pytest.raises(ValueError)),
            (tf.TensorShape((8, 64)), pytest.raises(ValueError)),
            (tf.TensorShape((8)), pytest.raises(ValueError)),
            (tf.TensorShape((8, 64, 64, 4, 5)), pytest.raises(ValueError)),
            (tf.TensorShape((8, 64, 64, 4, 5, 6)), pytest.raises(ValueError)),
        ],
    )
    def test_call_raises_error_wrong_input_spec(self, resblock_down, input_shape, expected_context):
        CommonLayerChecks.does_call_raises_error_wrong_input_spec(resblock_down, input_shape, expected_context)


class TestResBlockUp(DeepSakiLayerChecks):
    @pytest.fixture()
    def resblock_up(self):
        return ResBlockUp()

    @pytest.mark.parametrize(
        ("input_shape"),
        [
            tf.TensorShape((1, 16, 16, 4)),
            tf.TensorShape((8, 32, 32, 3)),
            tf.TensorShape((5, 32, 16, 12)),
            tf.TensorShape((3, 16, 32, 8)),
        ],
    )
    def test_call_correct_output_shape(self, input_shape):
        layer_instance = ResBlockUp()
        expected_shape = [
            input_shape[0],
            input_shape[1] * 2,
            input_shape[2] * 2,
            input_shape[3],
        ]
        CommonLayerChecks.has_call_correct_output_shape(layer_instance, input_shape, expected_shape)

    @pytest.mark.parametrize(
        ("input_shape", "expected_context"),
        [
            (tf.TensorShape((8, 64, 64, 4)), does_not_raise()),
            (tf.TensorShape((1, 32, 32, 3)), does_not_raise()),
            (tf.TensorShape((8, 64, 64)), pytest.raises(ValueError)),
            (tf.TensorShape((8, 64)), pytest.raises(ValueError)),
            (tf.TensorShape((8)), pytest.raises(ValueError)),
            (tf.TensorShape((8, 64, 64, 4, 5)), pytest.raises(ValueError)),
            (tf.TensorShape((8, 64, 64, 4, 5, 6)), pytest.raises(ValueError)),
        ],
    )
    def test_call_raises_error_wrong_input_spec(self, resblock_up, input_shape, expected_context):
        CommonLayerChecks.does_call_raises_error_wrong_input_spec(resblock_up, input_shape, expected_context)


class TestScaleLayer(DeepSakiLayerChecks):
    @pytest.fixture()
    def scale_layer(self):
        return ScaleLayer()

    @pytest.mark.parametrize(
        "input_shape",
        [
            tf.TensorShape([1]),
            tf.TensorShape([8, 16]),
            tf.TensorShape([8, 16, 16]),
            tf.TensorShape([8, 16, 16, 4]),
        ],
    )
    def test_call_correct_output_shape(self, scale_layer, input_shape):
        CommonLayerChecks.has_call_correct_output_shape(scale_layer, input_shape, input_shape)

    @pytest.mark.parametrize(
        ("input_shape", "expected_context"),
        [
            (tf.TensorShape((8)), does_not_raise()),
            (tf.TensorShape((8, 16)), does_not_raise()),
            (tf.TensorShape((8, 16, 16)), does_not_raise()),
            (tf.TensorShape((8, 16, 16, 4)), does_not_raise()),
            (tf.TensorShape((8, 16, 16, 4, 12)), does_not_raise()),
            (tf.TensorShape(()), pytest.raises(ValueError)),
        ],
    )
    def test_call_raises_error_wrong_input_spec(self, scale_layer, input_shape, expected_context):
        CommonLayerChecks.does_call_raises_error_wrong_input_spec(scale_layer, input_shape, expected_context)


class TestScalarGatedSelfAttention(DeepSakiLayerChecks):
    @pytest.fixture()
    def scalar_gated_self_attention(self):
        return ScalarGatedSelfAttention()

    @pytest.fixture(autouse=True)
    def mock_layers(self, mocker):
        calling_module = "DeepSaki.layers.layer_composites"
        # _mock_conv2d_block(mocker,calling_module) mock is not yet implemented
        _mock_scale_layer(mocker, calling_module)

    @pytest.mark.parametrize("use_spec_norm", [True, False])
    @pytest.mark.parametrize("intermediate_channel", [3, 7, 12, None])
    @pytest.mark.parametrize(
        "input_shape",
        [
            tf.TensorShape((1, 8, 8, 3)),
            tf.TensorShape((8, 8, 8, 4)),
            tf.TensorShape((1, 3, 3, 1)),
        ],
    )
    def test_call_correct_output_shape(self, input_shape, intermediate_channel, use_spec_norm):
        layer_instance = ScalarGatedSelfAttention(use_spec_norm, intermediate_channel)
        CommonLayerChecks.has_call_correct_output_shape(layer_instance, input_shape, input_shape)

    @pytest.mark.parametrize(
        ("input_shape", "expected_context"),
        [
            (tf.TensorShape((8, 64, 64, 4)), does_not_raise()),
            (tf.TensorShape((1, 32, 32, 3)), does_not_raise()),
            (tf.TensorShape((8, 64, 64)), pytest.raises(ValueError)),
            (tf.TensorShape((8, 64)), pytest.raises(ValueError)),
            (tf.TensorShape((8)), pytest.raises(ValueError)),
            (tf.TensorShape((8, 64, 64, 4, 5)), pytest.raises(ValueError)),
            (tf.TensorShape((8, 64, 64, 4, 5, 6)), pytest.raises(ValueError)),
        ],
    )
    def test_call_raises_error_wrong_input_spec(self, scalar_gated_self_attention, input_shape, expected_context):
        CommonLayerChecks.does_call_raises_error_wrong_input_spec(
            scalar_gated_self_attention, input_shape, expected_context
        )



import os

import pytest

os.environ["TF_CPP_MIN_LOG_LEVEL"] = "2"  # deactivate tensorflow warnings and infos. Keep Errors
import tensorflow as tf

from DeepSaki.losses.image_based_losses import ImageBasedLoss
from DeepSaki.losses.image_based_losses import PixelDistanceLoss
from DeepSaki.losses.image_based_losses import StructuralSimilarityLoss
from DeepSaki.losses.image_based_losses import LossCalcType
from DeepSaki.losses.image_based_losses import LossType


class TestImageBasedLossAbstractBase:
    @pytest.mark.parametrize(
        "calculation_type",
        [
            "Wrong String",
            3,
            5.6,
            None,
        ],
    )
    def test_init_raises_value_error(self, mocker, calculation_type):
        mocker.patch.multiple(ImageBasedLoss, __abstractmethods__=set())
        with pytest.raises(ValueError):
            _ = ImageBasedLoss(global_batch_size=64, calculation_type=calculation_type)

    @pytest.mark.parametrize(
        ("calculation_type", "expected"),
        [
            (LossCalcType.PER_CHANNEL, "_calc_loss_per_channel"),
            (LossCalcType.PER_IMAGE, "_calc_loss_per_image"),
        ],
    )
    def test_correct_loss_calc_func_selected(self, mocker, calculation_type, expected):
        # arange - Mock abstract class
        mocker.patch.multiple(ImageBasedLoss, __abstractmethods__=set())
        instance = ImageBasedLoss(global_batch_size=1, calculation_type=calculation_type)

        # hacky way of checking. reason: due to the mocking it is not possible to compare the functions
        assert instance.loss_calc_func.__name__ == expected

    @pytest.mark.parametrize(
        ("img1", "img2", "global_batch_size"),
        [
            (tf.ones(shape=(1, 1, 1, 4)), tf.ones(shape=(1, 1, 1, 4)), 1),
            (tf.ones(shape=(1, 1, 1, 3)), tf.ones(shape=(1, 1, 1, 3)), 64),
            (tf.ones(shape=(1, 64, 64, 4)), tf.ones(shape=(1, 64, 64, 4)), 64),
            (tf.ones(shape=(1, 10, 10, 4)), tf.ones(shape=(1, 10, 10, 4)), 50),
            (tf.ones(shape=(1, 99, 99, 3)), tf.ones(shape=(1, 99, 99, 3)), 87),
        ],
    )
    def test_calc_loss_per_image(self, mocker, img1, img2, global_batch_size):
        # arange - Mock abstract class and define error_func
        mocker.patch.multiple(ImageBasedLoss, __abstractmethods__=set())
        instance = ImageBasedLoss(global_batch_size=global_batch_size)
        error_func = lambda x, y: x * y * global_batch_size

        # act
        result = instance._calc_loss_per_image(img1, img2, error_func)
        expected = tf.ones_like(img1)

        # assert
        assert tf.math.reduce_all(result.numpy() == pytest.approx(expected.numpy(), 0.01))

    @pytest.mark.parametrize(
        ("img1", "img2", "global_batch_size", "channel_weights"),
        [
            (tf.ones(shape=(1, 1, 1, 4)), tf.ones(shape=(1, 1, 1, 4)), 1, [1, 1, 1, 3]),
            (tf.ones(shape=(1, 1, 1, 3)), tf.ones(shape=(1, 1, 1, 3)), 64, [1, 1, 2]),
            (tf.ones(shape=(1, 64, 64, 4)), tf.ones(shape=(1, 64, 64, 4)), 64, [1, 1, 1, 1]),
            (tf.ones(shape=(1, 10, 10, 4)), tf.ones(shape=(1, 10, 10, 4)), 50, [1, 1, 1, 3]),
            (tf.ones(shape=(1, 99, 99, 3)), tf.ones(shape=(1, 99, 99, 3)), 87, [1, 1, 2]),
        ],
    )
    def test_calc_loss_per_channel(self, mocker, img1, img2, global_batch_size, channel_weights):
        # arange - Mock abstract class and _get_channel_weights method and define error_func
        mocker.patch.multiple(ImageBasedLoss, __abstractmethods__=set())
        instance = ImageBasedLoss(global_batch_size=global_batch_size)
        mocker.patch(
            "DeepSaki.losses.image_based_losses.ImageBasedLoss._get_channel_weights", return_value=channel_weights
        )
        error_func = lambda x, y: x * y * global_batch_size

        # act
        result = instance._calc_loss_per_channel(img1, img2, error_func)
        expected = tf.reduce_sum(tf.ones_like(img1), axis=-1) + (channel_weights[-1] - 1)

        # assert
        assert tf.math.reduce_all(result.numpy() == pytest.approx(expected.numpy(), 0.01))

    @pytest.mark.parametrize(
        ("img_shape", "normalize_last_channel", "expected"),
        [
            (tf.TensorShape([1, 64, 64, 3]), False, [1, 1, 1]),
            (tf.TensorShape([1, 64, 64, 3]), True, [1, 1, 2]),
            (tf.TensorShape([1, 64, 64, 4]), False, [1, 1, 1, 1]),
            (tf.TensorShape([1, 64, 64, 4]), True, [1, 1, 1, 3]),
        ],
    )
    def test_get_channel_weights(self, mocker, img_shape, normalize_last_channel, expected):
        # arange - Mock abstract class
        mocker.patch.multiple(ImageBasedLoss, __abstractmethods__=set())
        instance = ImageBasedLoss(global_batch_size=1)
        # Act
        result = instance._get_channel_weights(img_shape, normalize_last_channel)
        # assert
        assert tf.math.reduce_all(result == pytest.approx(expected, 0.01))

    @pytest.mark.skip(reason="Not implemented yet.")
    def test_call(self):
        ...


class TestPixelDistanceLoss:
    @pytest.mark.parametrize(
        "loss_type",
        [
            "Wrong String",
            2,
            2.0,
            None,
        ],
    )
    def test_init_raises_error(self, loss_type):
        with pytest.raises(ValueError):
            _ = PixelDistanceLoss(global_batch_size=64, loss_type=loss_type)

    @pytest.mark.parametrize(
        ("loss_type", "expected"),
        [
            (LossType.MAE, tf.abs),
            (LossType.MSE, tf.square),
        ],
    )
    def test_init_correct_error_func_selected(self, loss_type, expected):
        loss = PixelDistanceLoss(global_batch_size=64, loss_type=loss_type)
        assert loss.error_func_type == expected

    @pytest.mark.parametrize(
        ("tensor1", "tensor2", "loss_type", "expected"),
        [
            (3 * tf.ones(shape=[1, 1, 1, 1]), tf.ones(shape=[1, 1, 1, 1]), LossType.MAE, tf.constant(2.0)),
            (3 * tf.ones(shape=[8, 64, 64, 4]), tf.ones(shape=[8, 64, 64, 4]), LossType.MAE, tf.constant(2.0)),
            (5 * tf.ones(shape=[1, 1, 1, 1]), tf.ones(shape=[1, 1, 1, 1]), LossType.MAE, tf.constant(4.0)),
            (5 * tf.ones(shape=[8, 64, 64, 4]), tf.ones(shape=[8, 64, 64, 4]), LossType.MAE, tf.constant(4.0)),
            (3 * tf.ones(shape=[1, 1, 1, 1]), tf.ones(shape=[1, 1, 1, 1]), LossType.MSE, tf.constant(4.0)),
            (3 * tf.ones(shape=[8, 64, 64, 4]), tf.ones(shape=[8, 64, 64, 4]), LossType.MSE, tf.constant(4.0)),
            (5 * tf.ones(shape=[1, 1, 1, 1]), tf.ones(shape=[1, 1, 1, 1]), LossType.MSE, tf.constant(16.0)),
            (5 * tf.ones(shape=[8, 64, 64, 4]), tf.ones(shape=[8, 64, 64, 4]), LossType.MSE, tf.constant(16.0)),
        ],
    )
    def test_error_func(self, tensor1, tensor2, loss_type, expected):
        loss = PixelDistanceLoss(global_batch_size=64, loss_type=loss_type)
        result = loss._error_func(tensor1, tensor2)
        assert tf.math.reduce_all(result.numpy() == pytest.approx(expected.numpy(), 0.01))

    def test_isinstance_tensorflow_loss(self):
        loss = PixelDistanceLoss(global_batch_size=64)
        assert isinstance(loss, tf.keras.losses.Loss)


class TestSsimLoss:
    @pytest.mark.parametrize(
        ("tensor1", "tensor2", "expected"),
        [
            (tf.ones(shape=[1, 1, 1, 1]), tf.ones(shape=[1, 1, 1, 1]), tf.constant(0.0)),
            (tf.ones(shape=[8, 64, 64, 4]), tf.ones(shape=[8, 64, 64, 4]), tf.constant(0.0)),
            (tf.ones(shape=[1, 1, 1, 1]), tf.zeros(shape=[1, 1, 1, 1]), tf.constant(1.0)),
            (tf.ones(shape=[8, 64, 64, 4]), tf.zeros(shape=[8, 64, 64, 4]), tf.constant(1.0)),
        ],
    )
    def test_error_func(self, tensor1, tensor2, expected):
        loss = StructuralSimilarityLoss(global_batch_size=64)
        result = loss._error_func(tensor1, tensor2)
        assert tf.math.reduce_all(result.numpy() == pytest.approx(expected.numpy(), 0.01))

    def test_isinstance_tensorflow_loss(self):
        loss = StructuralSimilarityLoss(global_batch_size=64)
        assert isinstance(loss, tf.keras.losses.Loss)



from typing import Any
from typing import Callable
from typing import Dict
from typing import Tuple

import tensorflow as tf

class ReflectionPadding2D(tf.keras.layers.Layer):
    """Reflection Padding layer with support for TPU.

    Info:
        At the time of implementation, Google's TPUs did not support gradient calculation of mirrored paddings like
        "SYMMETRIC" and "REFLECTION". Therefore, a custom gradient needed to be calculated, by cropping the padding and
        passing the gradients from the previous layer.
    """

    def __init__(self, padding: Tuple[int, int] = (1, 1), **kwargs: Any) -> None:
        """Initialize the `ReflectionPadding2D` layer.

        Args:
            padding (Tuple[int, int], optional): One-sided padding added to the `hight` and `width` to an input tensor
                of shape (batch, height, width, channel)respectively. Defaults to (1, 1).
            kwargs (Any): Additional key word arguments passed to the base class.
        """
        super(ReflectionPadding2D, self).__init__(**kwargs)
        self.padding = tuple(padding)
        self.input_spec = tf.keras.layers.InputSpec(ndim=4)

    @tf.custom_gradient
    def _padding_func(self, input_tensor: tf.Tensor) -> Tuple[tf.Tensor, Callable[[tf.Tensor], tf.Tensor]]:
        """Calculates results for forward prop and provides a function to calc. gradients for backward propagation.

        Args:
            input_tensor (tf.Tensor): Tensor of shape `(batch, height, width, channel)`.

        Returns:
            padded_tensor: Padded tensor of shape `(batch, height+2*padding, width+2*padding, channel)`.
            custom_grad: Function returning the gradients during backprop.
        """
        padding_height, padding_width = self.padding
        padding_tensor = [
            [0, 0],
            [padding_height, padding_height],
            [padding_width, padding_width],
            [0, 0],
        ]
        padded_tensor = tf.pad(input_tensor, padding_tensor, mode="REFLECT")

        # upstream gradient is the chainrule of all previous gradients!
        def custom_grad(upstream: tf.Tensor) -> tf.Tensor:
            # The gradients that represent the padding are cut, since they are not relevant!
            return tf.image.crop_to_bounding_box(
                image=upstream,
                offset_height=0,
                offset_width=0,
                target_height=upstream.shape[1] - 2 * padding_height,
                target_width=upstream.shape[2] - 2 * padding_width,
            )

        return padded_tensor, custom_grad

    def compute_output_shape(self, input_shape: tf.TensorShape) -> tf.TensorShape:
        """Calculates the expected output shape after calling the layer.

        Assumes "channels_last" configuration.

        Args:
            input_shape (tf.TensorShape): Shape of the input data to the layer.

        Returns:
            output_shape: expected output shape of the layer.
        """
        return tf.TensorShape(
            (
                input_shape[0],
                input_shape[1] + 2 * self.padding[0],
                input_shape[2] + 2 * self.padding[1],
                input_shape[3],
            )
        )

    def call(self, input_tensor: tf.Tensor) -> tf.Tensor:
        """Calls the `ReflectionPadding2D` layer.

        Args:
            input_tensor (tf.Tensor): Tensor of shape `(batch, height, width, channel)`.

        Returns:
            padded tensor of shape `(batch, height+2*padding, width+2*padding, channel)`.
        """
        return self._padding_func(input_tensor)

    def get_config(self) -> Dict[str, Any]:
        """Serialization of the object.

        Returns:
            Dictionary with the class' variable names as keys.
        """
        config = super(ReflectionPadding2D, self).get_config()
        config.update({"padding": self.padding})
        return config



from typing import Any
from typing import Dict

import tensorflow as tf

class ComplexActivation(tf.keras.layers.Layer):
    """Wrapper to apply a given `activation` to a complex input individually for the real and imaginary part.

    **Example:**
    ```python hl_lines="4"
    import DeepSaki as ds
    import tensorflow as tf

    activation = tf.keras.layers.LeakyReLU(alpha=0.3)

    complex_activation = ds.activations.ComplexActivation(activation=activation)
    complex_tensor = tf.complex(real=[-1.0,1.0], imag=[-2.0,2.0])
    x = complex_activation(complex_tensor)
    print(x)
    #output: <tf.Tensor: shape=(2,), dtype=complex64, numpy=array([-0.3-0.6j,  1. +2.j ], dtype=complex64)>
    ```
    """

    def __init__(self, activation: tf.keras.layers.Layer, **kwargs: Any) -> None:
        """Initialize ComplexActivation.

        Args:
            activation (tf.keras.layers.Layer): Activation function to complexyfy.
            kwargs: keyword arguments passed to the parent class tf.keras.layers.Layer.
        """
        super(ComplexActivation, self).__init__(**kwargs)
        self.activation = activation

    def call(self, inputs: tf.Tensor) -> tf.Tensor:
        """Splits its intput `inputs`into a real and imaginary part, applies `activation` and constructs a complex number.

        Args:
            inputs (tf.Tensor): Input tensor to be activated. Might be a complex or real valued tensor.

        Returns:
            Complex-valued tensor with activated real and imaginary part.
        """
        real = self.activation(tf.math.real(inputs))
        imag = self.activation(tf.math.imag(inputs))
        return tf.complex(real, imag)

    def get_config(self) -> Dict[str, Any]:
        """Returns configuration of class instance.

        Returns:
            Dict[str,Any]: Dictionary containing the class' configuration.
        """
        config = super(ComplexActivation, self).get_config()
        config.update({"activation": self.activation})
        return config



from typing import List
from typing import Optional
from typing import Tuple

import tensorflow as tf

from DeepSaki.initializers.he_alpha import HeAlphaNormal
from DeepSaki.initializers.he_alpha import HeAlphaUniform
from DeepSaki.layers.padding import ReflectionPadding2D
from DeepSaki.types.layers_enums import InitializerFunc
from DeepSaki.types.layers_enums import PaddingType

def get_initializer(
    initializer: InitializerFunc,
    seed: Optional[int] = None,
) -> tf.keras.initializers.Initializer:
    """Wrapper to return a certain initializer given a descriptive string.

    Args:
        initializer (InitializerFunc): Enum description of the initializer.
        seed (Optional[int], optional): Seed to make the behavior of the initializer deterministic. Defaults to None.

    Returns:
        Instance of an initializer object.
    """
    valid_options = {
        InitializerFunc.RANDOM_NORMAL: tf.keras.initializers.RandomNormal(mean=0.0, stddev=0.02, seed=seed),
        InitializerFunc.RANDOM_UNIFORM: tf.keras.initializers.RandomUniform(minval=-0.002, maxval=0.02, seed=seed),
        InitializerFunc.GLOROT_NORMAL: tf.keras.initializers.GlorotNormal(seed=seed),
        InitializerFunc.GLOROT_UNIFORM: tf.keras.initializers.GlorotUniform(seed=seed),
        InitializerFunc.HE_NORMAL: tf.keras.initializers.HeNormal(seed=seed),
        InitializerFunc.HE_UNIFORM: tf.keras.initializers.HeUniform(seed=seed),
        InitializerFunc.HE_ALPHA_NORMAL: HeAlphaNormal(seed=seed),
        InitializerFunc.HE_ALPHA_UNIFORM: HeAlphaUniform(seed=seed),
    }

    if initializer not in valid_options:
        raise ValueError(f"Undefined initializer provided: '{initializer}'. Valid options are: '{valid_options.keys()}")

    return valid_options.get(initializer)


def pad_func(
    pad_values: Tuple[int, int] = (1, 1),
    padding_type: PaddingType = PaddingType.ZERO,
) -> tf.keras.layers.Layer:
    """Wrapper to obtain a padding layer instance.

    Args:
        pad_values (Tuple[int,int], optional): Size of the padding values. Defaults to (1, 1).
        padding_type (PaddingType, optional): Padding Type. Defaults to PaddingType.ZERO.

    Returns:
        Instance of a padding layer object.
    """
    valid_options = {
        PaddingType.REFLECTION: ReflectionPadding2D(pad_values),
        PaddingType.ZERO: tf.keras.layers.ZeroPadding2D(pad_values),
    }
    if padding_type not in valid_options:
        raise ValueError(
            f"Undefined padding type provided: '{padding_type}'. Valid options are: '{valid_options.keys()}'"
        )

    return valid_options.get(padding_type)


def dropout_func(filters: int, dropout_rate: float) -> tf.keras.layers.Layer:
    """Wrapper to obtain a dropout layer depending on the number of filters of the preceeding feature map.

    Args:
        filters (int): Number of filters of the preceeding layer.
        dropout_rate (float): Probability of the dropout layer to drop weights.

    Returns:
        layer: Returns `tf.keras.layers.SpatialDropout2D` if number of filters > 1, otherwise 'tf.keras.layers.Dropout'.
    """
    if not isinstance(filters, int):
        raise TypeError(f"Parameter 'filters' shall be of type int but is: '{type(filters)}'")

    if filters > 1:
        return tf.keras.layers.SpatialDropout2D(dropout_rate)
    if filters == 1:
        return tf.keras.layers.Dropout(dropout_rate)
    raise ValueError(f"provided value '{filters}'for param 'filters' is unvalid. Provide an int bigger than 0.")


def plot_layer(
    layer: tf.keras.layers.Layer,
    input_shape: List[int],
) -> None:
    """Creates a model from a given layer to be able to call model.summary() and to plot a graph image.

    Args:
        layer (tf.keras.layers.Layer): Layer to be plotted.
        input_shape (List[int]): Shape of the desired input of the layer.
    """
    layer.build([None, *input_shape])
    inputs = tf.keras.layers.Input(shape=input_shape)
    model = tf.keras.Model(inputs=inputs, outputs=layer.call(inputs))
    model.summary()
    tf.keras.utils.plot_model(model, show_shapes=True, expand_nested=True, to_file=layer.name + ".png")



from DeepSaki import activations
from DeepSaki import augmentations
from DeepSaki import constraints
from DeepSaki import initializers
from DeepSaki import layers
from DeepSaki import losses
from DeepSaki import models
from DeepSaki import optimizers
from DeepSaki import types
from DeepSaki import utils

__version__ = "1.0.1"

__author__ = "Sascha Kirch"

__all__ = [
    "activations",
    "augmentations",
    "constraints",
    "initializers",
    "layers",
    "losses",
    "models",
    "optimizers",
    "types",
    "utils",
]



from DeepSaki.losses.image_based_losses import ImageBasedLoss
from DeepSaki.losses.image_based_losses import PixelDistanceLoss
from DeepSaki.losses.image_based_losses import StructuralSimilarityLoss

__all__ = [
    "ImageBasedLoss",
    "PixelDistanceLoss",
    "StructuralSimilarityLoss",
]
